# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted, FLOAT_DTYPES


def _check_length_scale(length_scale):
    """ We consider only isotropic length scales"""
    length = np.squeeze(length_scale)
    if np.ndim(length) != 0:
        raise ValueError('Only isotropic length steps are accepted')
    return length


def _check_has_gradient(kernel):
    names = ['dKdX1', 'dKdX2', 'hessian', 'hessian_diagonal']
    for name in names:
        if not hasattr(kernel, name):
            raise TypeError('Can add only kernels with implemented gradients')


class StandardScalerGrad(StandardScaler):
    """ Scale inputs and outputs derivatives as appropriate"""
    def transform(self, X, dy=None, copy=None):
        """ Perform standardization by centering and scaling.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            the data to be scaled

        dy : array-like of shape (n_samples, n_features) or None. Default None
            if not None, represents the output derivatives data that
            should be scaled

        copy: bool. Default None
            copy the input X or not

        Returns
        -------
        X_tr : array-like, shape (n_samples, n_features)
             Transformed array.

        dy_tr : if dy is not None, array-like, shape (n_samples, n_features)
             Transformed array.
        """
        if dy is None:
            return super().transform(X, copy)

        check_is_fitted(self)
        copy = copy if copy is not None else self.copy
        X_params = {'copy': copy, 'estimator': self,
                    'dtype': FLOAT_DTYPES,
                    'force_all_finite': 'allow-nan'}
        y_params = {'copy': copy, 'estimator': self,
                    'dtype': FLOAT_DTYPES,
                    'force_all_finite': 'allow-nan'}
        X, dy = self._validate_data(X, dy,
                                    validate_separately=(X_params,
                                                         y_params))
        if self.with_mean:
            X -= self.mean_
        if self.with_std:
            X /= self.scale_
            dy *= self.scale_
        return X, dy

    def inverse_transform(self, X=None, dy=None, copy=None):
        """ Perform standardization by centering and scaling.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features). Default None
            the data to be scaled

        dy : array-like of shape (n_samples, n_features) or None. Default None
            if not None, represents the output derivatives data that
            should be scaled

        copy: bool. Default None
            copy the input X or not

        Returns
        -------
        X_tr : array-like, shape (n_samples, n_features)
            Transformed array.

        dy_tr : if dy is not None, array-like, shape (n_samples, n_features)
             Transformed array.
        """
        if X is None and dy is None:
            raise ValueError('At least one of X or dy must be not None')

        if dy is None:
            return super().inverse_transform(X, copy)

        check_is_fitted(self)
        copy = copy if copy is not None else self.copy

        if X is not None:
            X = np.asarray(X)
        dy = np.asarray(dy)
        if copy:
            if X is not None:
                X = X.copy()
            dy = dy.copy()
        if self.with_std:
            if X is not None:
                X *= self.scale_
            dy /= self.scale_
        if self.with_mean:
            if X is not None:
                X += self.mean_
        if X is None:
            return dy
        return X, dy




# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np

import h5py

from clinamen import DESCRIPTORS_FLOAT_DTYPE as DTYPE


def write_descriptors(file_name, descriptors, descriptors_grads, ids,
                      name=None, flattened=True):
    """ Append new data to an existing dataset, if the dataset
    does not exist, create a new one

    Parameters
    ----------
    file_name : string
        the dataset name

    descriptors : 2D array-like of shape (n, d), if :attr:`flattened` is ``True``.
        n is the number of structures for which the descriptors
        were calculated. d is the dimensionality of the descriptors.
        If :attr:`flattened` is ``False``, descriptors can be a multidimensional
        array of shape (n, ...). 

    descriptors_grads : 2D array-like of shape (n, r) if :attr:`flattened` is ``True``.
        Otherwise, it can be a multidimensional array of shape (n, ...).
        It can also be ``None``. If not None, these are the (possibly flattened,
        if :attr:`flattened` is ``True``) Jacobians of the descriptors.

    name : string. Default None
        the system name. A tag that specifies the system when the
        dataset is created. If the dataset already exists, it checks
        the it corresponds to system ``name``

    ids : array-like of shape (n, )
        for each descriptor, is a string that identifies the structure
        coresponding to that descriptor
    """
    comp_kwargs = {'compression': 'gzip', 'compression_opts': 9}
    descriptors = np.atleast_2d(descriptors)
    if flattened:
        d = descriptors.shape[1]
        shape_x = (d, )
    else:
        d = descriptors.shape[1:]
        shape_x = d
    X = descriptors.copy()
    n_structs = X.shape[0]
    if n_structs != len(ids):
        raise ValueError(f'Have got descriptors for {n_structs} structures '
                         f'but {len(ids)} identifiers are given')
    shape_dx = tuple()
    if descriptors_grads is not None:
        descriptors_grads = np.atleast_2d(descriptors_grads)
        if flattened:
            r = descriptors_grads.shape[1]
            shape_dx = (r, )
        else:
            r = descriptors_grads.shape[1:]
            shape_dx = r
        DX_shape = descriptors_grads.shape
        DX = descriptors_grads.copy()
        if X.shape[0] != DX.shape[0]:
            raise ValueError(f'Have got descriptors for {X.shape[0]} '
                             f'structures but gradients for {DX.shape[0]}')
    new_dataset = True
    if os.path.isfile(file_name):
        new_dataset = False
        # get data to check dataset consistency
        with h5py.File(file_name, 'r') as f:
            sys_name = f['system'].attrs['name'].decode()
            existing_X_shape = tuple(f['system'].attrs['X_shape'])
            existing_DX_shape = tuple(f['system'].attrs['dX_shape'])
        if name is not None:
            c0 = (name == sys_name)
        else:
            c0 = True
        if not c0:
            raise ValueError(f'Dataset name tag {name} does not '
                             f'agree with the name tag {sys_name} '
                             f'of dataset {file_name}')
        if existing_X_shape != shape_x:
            raise ValueError(f'Descriptors have shape {shape_x}, but '
                             f'existing descriptors with shape '
                             f'{existing_X_shape} have been found in '
                             f'{file_name}')
        c1 = existing_DX_shape != shape_dx
        c2 = existing_DX_shape is not tuple()
        c3 = shape_dx is not tuple()
        if (c1 and c2) or (c1 and c3):
            raise ValueError(f'Descriptors Jacobians have shape {shape_dx}, '
                             f'but existing Jacobians with shape '
                             f'{existing_dX_shape} have been found in '
                             f'{file_name}')
    if new_dataset:
        with h5py.File(file_name, 'w') as f:
            system = f.create_group('system')
            if name is None:
                sys_name = 'system'
            else:
                sys_name = name
            system.attrs['name'] = np.string_(sys_name)
            system.attrs['X_shape'] = np.array(shape_x, dtype=np.uint16)
            system.attrs['dX_shape'] = np.array(shape_dx, dtype=np.uint16)
            dataX = f.create_group('X')
            dataX.attrs['name'] = np.string_('descriptors')
            for desc, struct_id in zip(X, ids):
                dataX.create_dataset(struct_id, data=desc, dtype=DTYPE,
                                     maxshape=shape_x, **comp_kwargs)
            if descriptors_grads is not None:
                dataDX = f.create_group('DX')
                dataDX.attrs['name'] = np.string_('descriptors_Jacobians')
                for desc_grad, struct_id in zip(DX, ids):
                    dataDX.create_dataset(struct_id, data=desc_grad,
                                          dtype=DTYPE,
                                          maxshape=shape_dx, **comp_kwargs)
    else:
        with h5py.File(file_name, 'a') as f:
            dataX = f['X']
            for desc, struct_id in zip(X, ids):
                if struct_id in dataX.keys():
                    continue
                dataX.create_dataset(struct_id, data=desc, dtype=DTYPE,
                                     maxshape=shape_x, **comp_kwargs)
            if descriptors_grads is not None:
                dataDX = f['DX']
                for desc_grad, struct_id in zip(DX, ids):
                    dataDX.create_dataset(struct_id, data=desc_grad,
                                          dtype=DTYPE,
                                          maxshape=shape_dx, **comp_kwargs)


def read_descriptors_by_id(file_name, ids):
    """ Given an id or a list thereof, it returns the eventual descriptors
    and Jacobians

    Parameters
    ----------
    file_name : string
        the hdf5 file from where the descriptors should be fetched

    ids : iterable
        the identity keys of the descriptors we want to fetch

    Returns
    -------
    X, DX indices: tuple
        the descriptors and their Jacobians, each as a list, and a list
        representing the indices corresponding to the ids in ``ids`` that
        were found.
        If the Jacobians are not present, None is returned
    """
    indices = []
    if not os.path.isfile(file_name):
        return None, None, []

    with h5py.File(file_name, 'r') as f:
        dataX = f['X']
        grads = True
        try:
            dataDX = f['DX']
        except KeyError:
            grads = False
        X_data = []
        DX_data = []
        for i, id_no in enumerate(ids):
            try:
                val = np.array(dataX[id_no], dtype=DTYPE)
            except KeyError:
                continue
            else:
                indices.append(i)
                X_data.append(val)
                if grads:
                    DX_data.append(np.array(dataDX[id_no], dtype=DTYPE))
        if grads:
            if len(X_data) != len(DX_data):
                raise ValueError(f'Found {len(X_data)} descriptors instances '
                                 f'but {len(DX_data)} Jacobians')
        if len(X_data) == 0:
            X_data = None
        if len(DX_data) == 0:
            DX_data = None
        return X_data, DX_data, indices



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

r""" Clinamen: Covariance-matrix-adaptation evolution strategy for Localized
Impurities with unsupervised Machine learning exploration of Non-elementary
structures.

**Clinamen** is a Python package implementing the covariance-matrix-adaptaion
evolution strategy (CMA-ES) for finding low-energy configurations of localized
defects. It additionally employs an unsupervised machine learning (ML) approach
for exploiting the structures visited by the evolutionary algorithm (EA) in order
to find additional low-energy minima.
"""
import numpy as np

from clinamen.__about__ import *

DESCRIPTORS_FLOAT_DTYPE = np.float32
INDIVIDUALS_FLOAT_DTYPE = np.float64



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import copy
import numpy as np
import pandas as pd
import os

from mpi4py import MPI

from ase import Atoms
from ase.io.vasp import write_vasp
from ase.calculators.calculator import CalculationFailed

from clinamen.evpd import ROOT_LOGGER as logger
from clinamen.evpd import LOGFILE_NAME
from clinamen.utils import counter

proc_no = MPI.COMM_WORLD.Get_rank()


class Individual(Atoms):
    """ Class for representing an individual in a population.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._total_energy_calculations = 0
        # used to keep track of changes in the structure
        self._atoms_check = None
        # keep track of the original structure
        self._original_structure = Atoms(*args, **kwargs)
        # For cloning instances
        self.calculator_factory = None
        self.calculator_parameters = None
        self.skip_calc = False  # when fitness if produced with metamodel
                                # it must be set to True to avoid another
                                # calculation

    def set_calculator_factory(self, calc_factory, calc_parameters):
        """
        Parameters
        ----------
        calc_factory : a class derived from
            ``ase.calculators.interface.Calculator`` or a function
            generating a calculator

        calc_parameters : dict
            keyword-value pairs used to initialize a ``calc_factory``
            instance
        """
        if callable(calc_factory) or issubclass(calc_factory, type):
            params = copy.deepcopy(calc_parameters)
            calculator = calc_factory(**calc_parameters)
            calculator.calculator_parameters = copy.deepcopy(params)
            calculator.calculator_results = dict()
            self.calculator_factory = calc_factory
            self.calculator_parameters = copy.deepcopy(params)
            super().set_calculator(calculator)
        else:
            raise TypeError('calc_factory must be an object able to '
                            'instantiate a new calculator, i.e. '
                            'either a function or a class')

    @counter
    def get_potential_energy(self, *args, **kwargs):
        attempt_count = 0
        while True:
            try:
                attempt_count += 1
                if proc_no == 0:
                    logger.debug(f'Process {proc_no}: '
                                 'Attempting energy calculation for '
                                 f'individual: {self.my_name}\n'
                                 f'Attempt number: {attempt_count}')
                ene = super().get_potential_energy(*args, **kwargs)
                if proc_no == 0:
                    logger.debug(f'Process {proc_no}: '
                                 f'Energy counter +1 for "{self.my_name}"')

            except CalculationFailed:
                if proc_no == 0:
                    logger.debug(f'Process {proc_no}: '
                                 f'Calculating fitness for '
                                 f'individual {self.my_name} '
                                 'failed. Trying again...')
                continue
            else:
                self._total_energy_calculations += 1
                break
        self.calc.calculator_results['energy'] = ene
        # update
        self._atoms_check = self.copy()
        return ene

    def get_forces(self, *args, **kwargs):
        forces = super().get_forces(*args, **kwargs)
        self.calc.calculator_results['forces'] = forces.copy()
        return forces

    def calculation_required(self):
        """ Returns True if a new energy calculation is required """
        if not self.calc:
            raise RuntimeError('The Individual instance has no '
                               'set calculator')
        try:
            self.calc.calculator_results['energy']
        except KeyError:    # energy was never calculated
            return True
        changes = ['positions', 'cell', 'numbers', 'pbc']
        if self.skip_calc:
            return False
        else:
            for prop in changes:
                value1 = getattr(self, prop)
                value2 = getattr(self._atoms_check, prop)
                if not (value1 == value2).all():
                    return True

    def update_fitness(self):
        """ Calculate the total energy of the individual """
        if proc_no == 0:
            logger.debug(f'Process {proc_no}: '
                         'Calculating fitness of '
                         f'individual "{self.my_name}"')
        if self.calc is not None:
            if self.calculation_required():
                self._fitness = -self.get_potential_energy()
            else:
                if proc_no == 0:
                    logger.debug(f'Process {proc_no}: '
                                 f'{self.my_name} did not change. Using '
                                 'cached fitness')
                self._fitness = -self.calc.calculator_results['energy']
        else:
            if proc_no == 0:
                logger.critical(f'Process {proc_no}: '
                                'Trying to calculate the cost, '
                                f'but the Individual "{self.my_name}" '
                                'instance has no calculator')
            raise RuntimeError('The Individual instance has no '
                               'set calculator')

    @property
    def cost(self):
        """ Value of the cost function of the individual (its energy) """
        return -self.fitness

    @property
    def fitness(self):
        """ Fitness value of the individual """
        self.update_fitness()  # this will initialize self._fitness
        if hasattr(self, '_fitness_transformed'):
            return self._fitness_transformed
        return self._fitness

    @property
    def defect_position(self):
        """ Location of the eventual defect in the structure """
        return self._defect_position

    @defect_position.setter
    def defect_position(self, value):
        """ Scaled positions of the defective site """
        if isinstance(value, (list, tuple, np.ndarray)):
            self._defect_position = np.array(value)
        else:
            if proc_no == 0:
                logger.exception(f'Process {proc_no}: '
                                 'The defect position is required, but '
                                 f'the Individual "{self.my_name}" '
                                 'instance has no set defect position.')
            raise TypeError('Defect position must be given as a 1D array')

    @property
    def metric_tensor(self):
        """ The metric tensor of the cell """
        cell = self.get_cell()[:]
        metric = np.dot(cell, cell.T)
        return metric

    @property
    def distances_from_defect(self):
        """ Distance of each atom in the system from the defect """
        self._distances_from_defect = self._get_distances_from_defect()
        return self._distances_from_defect
    
    @property
    def my_name(self):
        """ Identifier """
        if not hasattr(self, '_my_name'):
            self._my_name = 'individual'
        return self._my_name

    @my_name.setter
    def my_name(self, value):
        self._my_name = value

    @property
    def etol(self):
        """ Tolerance for comparing costs between two individuals """
        return self._etol

    @etol.setter
    def etol(self, value):
        self._etol = value

    @property
    def drel(self):
        r""" Tolerance for the relative distance between two individuals.
        The relative distance is defined as:

        .. math::

            d_{rel}(i, j) = \frac{\sum_k |d_i(k) - d_j(k)|}{\sum_k d_i(k)}
            
        """
        return self._drel

    @drel.setter
    def drel(self, value):
        self._drel = value

    @property
    def dmax(self):
        """ Tolerance for the maximum distance discrepancy
        between two individuals.
        This distance is defined as:

        .. math::

            d_max(i, j) = max_k(|d_i(k) - d_j(k)|)
            
        """
        return self._dmax

    @dmax.setter
    def dmax(self, value):
        self._dmax = value

    @property
    def dmin(self):
        """ Tolerance for the minimum distance at which two atoms
        can be located. Used to reject a structure where two atoms
        are too close. 

        Default value = 0.25 minimum bond length in the system
        """
        if hasattr(self, '_dmin') and self._dmin is not None:
            return self._dmin
        distances = self._original_structure.get_all_distances(mic=True)
        mindist = distances[0][1:]
        mindist = np.min(mindist)
        self._dmin = 0.25*mindist
        return self._dmin

    @dmin.setter
    def dmin(self, value):
        self._dmin = value

    @property
    def chromosome(self):
        """ The chromosome of an individual is the set of displacements
        from an initial configuration. Usually one very similar to the
        pristine system
        """
        chromosome = self.positions - self._original_structure.positions
        return chromosome

    @property
    def total_energy_calculations(self):
        """ Number of times the energy of the individual has been calculated
        """
        return self._total_energy_calculations

    @property
    def data(self):
        positions = [['x' + str(i), 'y' + str(i), 'z' + str(i)]
                     for i in range(len(self))]
        positions = [tag for sublist in positions for tag in sublist]
        data_pos = self.positions.ravel()
        displacements = [['dx' + str(i), 'dy' + str(i), 'dz' + str(i)]
                         for i in range(len(self))]
        displacements = [tag for sublist in displacements for tag in sublist]
        try:
            data_disp = self.chromosome.ravel()
        except AttributeError:
            data_disp = np.zeros_like(data_pos)
        lpos = ['Positions']*len(positions)
        lchrom = ['Chromosome']*len(displacements) 
        level1 = lpos + lchrom
        level1.extend(['Fitness']*2)
        level2 = positions + displacements
        level2.extend(['From Energy', 'Actual'])
        index = pd.MultiIndex.from_arrays([level1, level2])
        data = list(data_pos)
        data.extend(list(data_disp))
        if not self.calculation_required():
            transformed_fitness = self.fitness #update fitness if needed
            from_energy = self._fitness
            data.extend([from_energy, transformed_fitness])
        else:
            data.extend([np.nan, np.nan])
        series = pd.Series(index=index, data=data, dtype=np.float,
                           name=self.my_name)
        return series

    @property
    def founder(self):
        if hasattr(self, '_founder'):
            return self._founder
        founder_atoms = self._original_structure
        founder = Individual.make_individual_from_ase_atoms(founder_atoms)
        if hasattr(self, '_defect_position'):
            founder.defect_position = self.defect_position.copy()
        founder.my_name = self.my_name + '_founder'
        if hasattr(self, '_etol'):
            founder.etol = self.etol
        if hasattr(self, '_drel'):
            founder.drel = self.drel
        if hasattr(self, '_dmax'):
            founder.dmax = self.dmax
        if hasattr(self, '_dmin'):
            founder.dmin = self.dmin
        founder._original_structure = self._original_structure.copy()
        if proc_no == 0:
            logger.debug(f'Process {proc_no}: '
                         f'Founder  "{founder.my_name}" created')
        self._founder = founder
        return self._founder

    def has_proper_structure(self):
        """ Returns False if any interatomic distance in the system is
        smaller than ``self.dmin``.
        """
        distances = self.get_all_distances(mic=True)
        #remove diagonal elements
        vdistances = distances[~np.eye(distances.shape[0],dtype=bool)]
        return (vdistances > self.dmin).all()

    def _get_distances_from_defect(self):
        """ Calculate the distance of each atom in the cell from the defect
        site using PBCs """
        def_pos = self.defect_position
        positions = self.get_scaled_positions()
        cell = self.get_cell()[:]
        diff_pos = positions - def_pos
        diff_pos -= np.rint(diff_pos)
        metric = self.metric_tensor
        distances = np.diag(np.dot(diff_pos, np.dot(metric.T, diff_pos.T)))
        return np.sqrt(distances)

    def _get_distances_from_origin(self):
        scaled_pos = self.get_scaled_positions()
        scaled_pos -= np.rint(scaled_pos)
        metric = self.metric_tensor
        distances = np.diag(np.dot(scaled_pos, np.dot(metric.T, scaled_pos.T)))
        return np.sqrt(distances)

    def write_poscar(self, path=None):
        """ If path is not None, it is the folder where the POSCAR file
        will be saved
        """
        filename = 'POSCAR_' + self.my_name
        if path:
            filename = os.path.join(path, filename)
        write_vasp(filename, self, direct=True, sort=None, vasp5=True)
        if proc_no == 0:
            logger.info(f'Process {proc_no}: '
                        f'Individual "{self.my_name}" configuration '
                        'saved as POSCAR file '
                        f'with name: "{filename}"')

    def clone(self):
        """ Copy the instance, return a new instance """
        if proc_no == 0:
            logger.debug(f'Process {proc_no}: '
                         f'Creating clone of Individual "{self.my_name}"')
        copied_item = Individual.make_individual_from_ase_atoms(self)
        copied_atoms = self.copy()
        copied_original = self._original_structure.copy()
        copied_item._original_structure = copied_original
        copied_item._atoms_check = copied_atoms
        copied_item.set_calculator_factory(self.calculator_factory,
                                           self.calculator_parameters)
        if hasattr(self, '_fitness'):
            copied_item.calc.calculator_results['energy'] = \
                self.calc.calculator_results['energy']
            try:
                copied_item.calc.calculator_results['forces'] = \
                    self.calc.calculator_results['forces'].copy()
            except KeyError:
                pass
        if hasattr(self, '_defect_position'):
            copied_item.defect_position = self.defect_position.copy()
        copied_item.my_name = self.my_name + '_clone'
        if hasattr(self, '_etol'):
            copied_item.etol = self.etol
        if hasattr(self, '_drel'):
            copied_item.drel = self.drel
        if hasattr(self, '_dmax'):
            copied_item.dmax = self.dmax
        if hasattr(self, '_dmin'):
            copied_item.dmin = self.dmin
        if hasattr(self, '_fitness'):
            copied_item._fitness = self._fitness
        if hasattr(self, '_fitness_transformed'):
            copied_item._fitness_transformed = self._fitness_transformed
        if proc_no == 0:
            logger.debug(f'Process {proc_no}: '
                         f'Clone  "{copied_item.my_name}" created')
        return copied_item

    def calculate_comparison_distances(self, other):
        """ Calculates the relative distance and the maximum distance
        discrepancy between this individual and another one.

        Parameters
        ----------
        other : Individual instance

        Return
        ------
        rel_dist, max_dist : the relative and maximum discrepancy 
            distance between the two instances
        """
        d_1 = self._get_distances_from_origin()
        d_2 = other._get_distances_from_origin()
        diff = np.abs(d_1 - d_2)
        num = diff.sum()
        den = np.sum(d_1)
        return num/den, np.max(diff)

    def optimize_structure(self, **kwargs):
        """ Optimize the structure

        Parameters
        ----------
        kwargs : dict
            parameters for running the geometry optimization.
            A mandatory key is `optimizer`, which is an ase optimizer.
            The other key:value pairs are the parameters to supply to
            `optimizer`.
        """
        if proc_no == 0:
            logger.info(f'Process {proc_no}: '
                        'Optimizing structure of '
                        f'individual "{self.my_name}"')
        optimizer = kwargs.pop('optimizer', None)
        if optimizer is None:
            raise ValueError('An optimizer should be given for '
                             'optimizing the structure')
        with open(LOGFILE_NAME, 'a') as fileobj:
            optimizer = optimizer(self, logfile=fileobj)
            optimizer.run(**kwargs)
        if proc_no == 0:
            logger.info(f'Process {proc_no}: Optimization completed')

    @staticmethod
    def make_individual_from_ase_atoms(atoms):
        """ Takes an ase.Atoms instance and returns
        an evpd.core.Individual instance.

        The result is analogous to using ``atoms.copy()``,
        but also the calculator will be copied.
        """
        if not isinstance(atoms, Atoms):
            raise TypeError('The function argument is not '
                            'an ase.Atoms instance')
        if proc_no == 0:
            logger.debug(f'Process {proc_no}: '
                         'Generating an Individual from '
                         'an Atoms object')
        cell = atoms.cell[:].copy()
        pbc = atoms.pbc.copy()
        celldisp = atoms._celldisp.copy()
        info = atoms.info.copy()
        individual = Individual(cell=cell, pbc=pbc,
                                celldisp=celldisp, info=info)
        individual.arrays = dict()
        for key, value in atoms.arrays.items():
            individual.arrays[key] = value.copy()
        individual.constraints = copy.deepcopy(atoms.constraints)

        individual._original_structure = atoms.copy()

        return individual



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
import os
import pandas as pd
import scipy.stats
import matplotlib.pyplot as plt
import h5py
from mpi4py import MPI

from clinamen.evpd.core.individual import Individual
from clinamen.evpd.core.population import Population
from clinamen.cmaes.evolution import (StrategyParameters, TerminationCriteria,
                                      CMAES, GpCMAES)
from clinamen.cmaes.fitness_calculators import *


class PopulationEvolver:
    """ Evolves a ``Population`` instance using the CMA-ES algorithm

    Parameters
    ----------
    founder : ``evpd.core.individual`` instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    step_size : float > 0
        initial step size used in the CMA-ES algorithm
        default 0.2 Angstrom

    covariance : None or float or 1D array or 2D array
        the initial covariance matrix. Default is the identity matrix.
        If ``covariance`` is a float, then the matrix is diagonal with that
        value on the diagonal. If it is 1D array, it is still diagonal
        with that array on the diagonal.

    dmin : float
        minimum distance between two atoms to consider an individual to
        be valid.
        Default None (0.5 of the minimum bond distance)

    random_seed : int
        random seed to be used for generating random variates.
        Default to 10
    """

    cmaes_obj = CMAES
    fitness_calc_obj = FitnessCalculator

    def __init__(self, founder, step_size=0.2, covariance=None,
                 dmin=None, random_seed=10):
        if not isinstance(founder, Individual):
            raise TypeError("The founder must be an instance of "
                            "'evpd.core.individual.Individual'")
        self.founder = founder.clone()
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug(f'Individual "{self.founder.my_name}" '
                         'renamed as "Founder"')
        self.founder.my_name = 'Founder'
        self.founder.dmin = dmin

        if not founder.has_proper_structure():
            raise ValueError('The Founder has an improper structure: '
                             'probably two atoms are too close together '
                             '(minimum distance between two atoms must be '
                             '> {:.3f} Angstrom). Produce a new Founder.'.format(
                                 self.founder.dmin))

        dimension = len(self.founder.positions.ravel())
        self._dimension = dimension

        if covariance is None:
            self._covariance = np.eye(dimension)
        elif isinstance(covariance, (int, float)):
            self._covariance = np.eye(dimension)*covariance
        else:
            covariance = np.array(covariance)
            if covariance.ndim == 1:
                self._covariance = np.diag(covariance)
            elif covariance.ndim == 2:
                self._covariance = covariance
            else:
                raise TypeError('The input covariance matrix must be a '
                                '2D array')

        self._population = Population(self.founder)
        self._strategyparameters = StrategyParameters(dimension)
        self._terminationcriteria = TerminationCriteria()

        self._random_seed = random_seed
        self._step_size = step_size

        self._cmaes_params = {'strategy_params': self._strategyparameters,
                              'terminator': self._terminationcriteria,
                              'mean': self.founder.positions.ravel(),
                              'step_size': self._step_size,
                              'covariance': self._covariance.copy(),
                              'random_seed': self._random_seed}

        self._external_cmaes = False    # True if the method set_cmaes is used
        self._fitness_calculator = None
        # hdf5 file which keeps track of all individuals in the population
        # along the whole evolution. The file only keeps minimal information
        self.evolution_history = 'evolution_history.hdf5'
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug(f'Saving population data for initialized population.')
            self.save_population(0)

    @property
    def cmaes_parameters(self):
        """ Return a dictionary with the objects needed to initialize the
        instance's CMAES object"""
        return self._cmaes_params

    @property
    def population(self):
        """ The current ``Population`` instance """
        return self._population

    @property
    def fitness_calculator(self):
        return self._fitness_calculator

    def set_strategy_parameters(self, strategy_params):
        """ Set the values of the strategy parameters to overwrite the
        default ones.

        Parameters
        ----------
        strategy_params : instance of ``StrategyParameters``
        """
        self._strategyparameters = strategy_params
        self._cmaes_params['strategy_params'] = self._strategyparameters

    def set_termination_criteria(self, termination_criteria):
        """ Set the values of the termination criteria to overwrite the
        default ones.

        Parameters
        ----------
        termination_criteria : instance of ``TerminationCriteria``
        """
        self._terminationcriteria = termination_criteria
        self._cmaes_params['terminator'] = self._terminationcriteria

    def _check_cmaes_type(self, cmaes):
        if not type(cmaes) is self.cmaes_obj:
            raise TypeError('Loaded a {} but a {} instance '
                            'is required'.format(type(cmaes),
                                                 type(self.cmaes_obj)))

    def set_cmaes(self, cmaes):
        """ Set the a custom ``CMAES`` object to overwrite the
        default one.

        Parameters
        ----------
        cmaes : instance of ``CMAES``
        """
        self._check_cmaes_type(cmaes)
        self._cmaes = cmaes
        self._cmaes_params = {'strategy_params': cmaes.StrategyParameters,
                              'terminator': cmaes.Terminator,
                              'mean': cmaes.m,
                              'covariance': cmaes.C,
                              'step_size': cmaes.step_size,
                              'random_seed': cmaes.random_seed}
        self._external_cmaes = True
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.info('An external CMAES object has been set.\n'
                        'If it is a restart, the algorithm will restart from '
                        f'generation {cmaes.g}')

    @property
    def cmaes(self):
        return self._cmaes

    def _mutate_individual(self, individual):
        x_ind = self.cmaes._mutate_single().reshape(-1, 3)
        individual.positions = x_ind

    def _generate_mutated_population(self):
        max_iter = 1000
        pop = []
        trial_no = 0
        for i in range(self.cmaes.pop_size):
            ind = self.founder.clone()
            ind.my_name = 'Individual_{}_{}'.format(self.cmaes.g + 1, i)
            self._mutate_individual(ind)
            while not ind.has_proper_structure():
                self._mutate_individual(ind)
                trial_no += 1
                if trial_no > max_iter:
                    sigma = self.cmaes.step_size
                    if isinstance(self, GpPopulationEvolver):
                        c_a = self.cmaes.gradient_coefficient
                    else:
                        c_a = 0
                    raise RuntimeError(
                        'Mutation cannot generate a valid '
                        'structure after {} trials. '
                        'You are probably using a too large step_size '
                        '(current value = {:.3f}) or a too large '
                        'gradient coefficient (current value = {:.3f})'
                        .format(max_iter, sigma, c_a))
            pop.append(ind)
        return Population(pop)

    def get_object_parameters(self):
        """ Returns the object parameters as a NumPy 2D array of shape
        (N, d), where N is the number of individuals in the population and
        d is the search space dimension
        """
        pop = self.population
        size = (self.cmaes.pop_size, self._dimension)
        x = np.empty(size) 
        for i, ind in enumerate(pop):
            x[i] = ind.positions.flatten()
        return x

    def _get_fitness_calculator(self):
        return self.fitness_calc_obj(self.population)

    def _set_up_cmaes(self):
        if not self._external_cmaes:
            self._cmaes = self.cmaes_obj(**self._cmaes_params)

    def _evolve(self):
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug(f'Mutating individuals for generation '
                         f'{self.cmaes.g + 1}')
        pop = self._generate_mutated_population()
        self._population = pop
        self._fitness_calculator.population = self.population
        self.cmaes.set_fitness_calculator(self.fitness_calculator)
        self.cmaes.set_mutated_offspring(self.get_object_parameters())
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug('Mutation completed.')

    def _initialize_evolution(self):
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug('Initializing evolution: setting up the CMAES')
        self._set_up_cmaes()
        self._fitness_calculator = self._get_fitness_calculator()
        self._evolve()

    def evolve_population(self):
        """ Evolve the current population.
        Returns a generator with the relevant parameters of the
        current generation.
        """
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.info('Initializing the evolution process')
        self._initialize_evolution()
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.info('Initialization completed, commence the evolution')
        for params in self.cmaes.evolve(manual_mutation=True):
            if MPI.COMM_WORLD.Get_rank() == 0:
                g = params['g']
                logger.info(f'COMPLETED CALCULATIONS FOR GENERATION {g}.')
                logger.debug(f'Saving population data for generation {g}.')
                self.save_population(g)
            yield params
            if MPI.COMM_WORLD.Get_rank() == 0:
                logger.debug('Initializing next generation.')
            self._evolve()

    def save_population(self, generation):
        """ Append the current population to ``self.evolution_history``
        file

        Parameters
        ----------
        generation : int
            the index of the current generation. Used to create a
            corresponding new group in the hdf5 file
        """
        # to be consistent with the ids created by the function in evpd.misc
        df = '<f8'
        di = '<u1'
        no_dec = 10
        comp_kwargs = {'compression': 'gzip', 'compression_opts': 9}
        if not os.path.isfile(self.evolution_history):
            logger.debug('Creating new history file '
                         f'{self.evolution_history}')
            cell = self.founder.cell[:].copy()
            cell = np.array(cell, dtype=df).round(no_dec)
            numbers = self.founder.numbers.copy()
            numbers = np.array(numbers, dtype=di)
            pbc = self.founder.pbc.copy()
            pbc = np.array(pbc, dtype=di)

            with h5py.File(self.evolution_history, 'w') as f:
                group = f.create_group('system')
                group.create_dataset('cell', data=cell, shape=(3, 3),
                                     dtype=df, **comp_kwargs)
                group.create_dataset('atomic_numbers', data=numbers,
                                     shape=(len(numbers), ), dtype=di,
                                     **comp_kwargs)
                group.create_dataset('pbc', data=pbc, shape=(3, ),
                                     dtype=df, **comp_kwargs)

        d = self.population[0].positions.ravel().shape[0]
        shape = (len(self.population), d)
        names = [x.my_name for x in self.population]
        names = np.array(names, dtype='<S30')
        positions = []
        for ind in self.population:
            pos = ind.positions.ravel().astype(df).round(no_dec)
            positions.append(pos)
        positions = np.vstack(positions)
        logger.debug(f'Adding {positions.shape[0]} new instances of '
                     f'generation {generation} to history '
                     f'file {self.evolution_history}.')
        with h5py.File(self.evolution_history, 'a') as f:
            try:
                group = f.create_group('Generation_' + str(generation))
                group.create_dataset('labels', data=names,
                                     shape=(len(names), ),
                                     dtype='<S30', **comp_kwargs)
                group.create_dataset('positions', data=positions, shape=shape,
                                     dtype=df, **comp_kwargs)
            except ValueError:  # Unable to create group (name already exists)
                logger.debug(f'Population data for generation {generation} '
                             'are already present in the dataset.')


class SSPopulationEvolver(PopulationEvolver):
    """ Add to the initial covariance matrix a rank-s matrix increasing the
    variance for coordinates representing atoms close to the point defect
    position

    Parameters
    ----------
    nn_cutoff : float
        cutoff radius, in Angstrom, including the atoms used to build the
        rank-s matrix

    c_r : float
        coefficient to control the contribution of the rank-s matrix to
        the initial covariance matrix

    founder : evpd.core.individual instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    kwargs : other keyword arguments necessary to initialize a ``PopulationEvolver``
        instance
    """
    def __init__(self, nn_cutoff, c_r, founder, **kwargs):
        super().__init__(founder, **kwargs)
        self._nn_cutoff = nn_cutoff
        self._cr = c_r

    @property
    def number_of_nn(self):
        """ The number of nearest neighbors atoms within ``self.nn_cutoff``
        """
        if not hasattr(self, '_number_nn'):
            self._find_subspace_basis()
        return self._number_nn

    @property
    def nn_cutoff(self):
        """ Cutoff selecting the NN to the defect which will form the basis
        for the selected subspace
        """
        return self._nn_cutoff

    @property
    def basis_coefficients(self):
        """ The distances of the atoms within the cutoff and the coefficients
        per atom which are used to add the rank-s matrix
        to the initial covariance matrix
        """
        return (self._sorted_distances_within.copy(),
                self._basis_coefficients[::3].copy())

    @property
    def selected_subspace_basis(self):
        """ The basis spanning the selected subspace. Order with respect to
        the atomic distances from the defect
        """
        if not hasattr(self, '_number_nn'):
            self._find_subspace_basis()
        return self._subspace_basis.copy()

    @property
    def atoms_within_cutoff(self):
        """ Indices of the atoms within the cutoff """
        if not hasattr(self, '_number_nn'):
            self._find_subspace_basis()
        return self._atoms_within_cutoff

    def _find_subspace_basis(self):
        """ Find the atoms within ``self.number_of_nn`` from
        the defect and use their coordinates (plus those of the defect,
        if it is an atom) to generate the selected-subspace basis
        """
        cutoff = self.nn_cutoff
        founder = self.founder
        d_from_def = founder.distances_from_defect
        def_pos = founder.defect_position
        indices = np.argsort(d_from_def)
        sorted_distances = d_from_def[indices]
        self._Rmax = sorted_distances[-1]
        within_cf = sorted_distances <= cutoff
        within_atoms = indices[within_cf]
        self._Rc = sorted_distances[len(within_atoms) - 1]
        if self._Rc > self._Rmax:    # e.g. if the user chooses a cutoff larger
            self._Rc = self._Rmax    # than the supercell
        self._sorted_distances_within = sorted_distances[within_cf]
        has_def = (np.abs(founder.get_scaled_positions() - def_pos) < 1e-3)
        has_def = has_def.all(axis=1).any()
        if has_def:
            self._number_nn = len(np.unique(sorted_distances[within_cf])) - 1
            self._dnn = sorted_distances[1]
        else:
            self._number_nn = len(np.unique(sorted_distances[within_cf]))
            self._dnn = sorted_distances[0]
        self._atoms_within_cutoff = within_atoms
        m = len(self._atoms_within_cutoff)
        # basis vectors as row vectors
        basis = np.zeros((3*m, self._dimension))
        for i, atom in enumerate(self._atoms_within_cutoff):
            basis[3*i:3*i + 3, 3*atom:3*atom + 3] = np.eye(3)
        self._subspace_basis = basis

    def _calculate_basis_coefficients(self):
        d = self._sorted_distances_within
        betas = 1/((1 + d)**2)
        weights = np.repeat(betas, 3)
        self._basis_coefficients = weights*weights

    def _calculate_rank_s_matrix(self):
        #omag = np.mean(np.diag(self._covariance))    # magnitude of cov
        omag = 1
        basis = self.selected_subspace_basis.T
        self._calculate_basis_coefficients()
        basis *= np.sqrt(self._basis_coefficients)
        matrix = np.dot(basis, basis.T)*omag
        return matrix

    def _set_up_cmaes(self):
        if not self._external_cmaes:
            matrix = self._calculate_rank_s_matrix()*np.square(self._cr)
            C_0 = self._covariance
            self._cmaes_params['covariance'] = C_0 + matrix
        super()._set_up_cmaes()


class GpPopulationEvolver(SSPopulationEvolver):
    """ Exploit the gradient during the run.

    Parameters
    ----------
    c_alpha : float
        coefficient describing the relevance of the gradient term

    nn_cutoff : float
        cutoff radius, in Angstrom, including the atoms used to build the
        rank-s matrix

    c_r : float
        coefficient to control the contribution of the rank-s matrix to
        the initial covariance matrix

    founder : evpd.core.individual instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    kwargs : other keyword arguments necessary to initialize a ``PopulationEvolver``
        instance
    """

    cmaes_obj = GpCMAES
    fitness_calc_obj = FitnessGradientCalculator

    def __init__(self, c_alpha, nn_cutoff, c_r, founder, **kwargs):
        super().__init__(nn_cutoff, c_r, founder, **kwargs)
        self._c_alpha = c_alpha

    @property
    def gradient_coefficient(self):
        return self._c_alpha

    def _set_up_cmaes(self):
        if not self._external_cmaes:
            matrix = self._calculate_rank_s_matrix() * np.square(self._cr)
            C_0 = self._covariance
            self._cmaes_params['covariance'] = C_0 + matrix
            self._cmaes = self.cmaes_obj(**self._cmaes_params)
            self.cmaes.set_gradient_coefficient(self._c_alpha)
        else:
            self._c_alpha = self.cmaes.gradient_coefficient


class RSPopulationEvolver(SSPopulationEvolver):
    """ Restricted-subspace population evolver: only the genotype for
    atoms inside a cutoff radius is considered

    Parameters
    ----------
    nn_cutoff : float
        cutoff radius, in Angstrom, including the atoms used to build the
        rank-s matrix

    c_r : float
        coefficient to control the contribution of the rank-s matrix to
        the initial covariance matrix

    founder : evpd.core.individual instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    kwargs : other keyword arguments necessary to initialize a ``PopulationEvolver``
        instance


    Notes
    -----
    Similar to ``SSPopulationEvolver`` but only the atoms within the cutoff
    are moved.
    """

    fitness_calc_obj = RSFitnessCalculator

    @property
    def use_reduced_population_size(self):
        """ Bool, if True, choose automatically the population size
        as based on the dimension of the restricted subspace. If False,
        uses the population size given by the ``StrategyParameters``
        instance given at initialization. Default False.
        """
        return self._use_reduced_pop_size

    @use_reduced_population_size.setter
    def use_reduced_population_size(self, value):
        self._use_reduced_pop_size = value

    def get_object_parameters(self):
        """ Returns the object parameters as a NumPy 2D array of shape
        (N, d), where N is the number of individuals in the population and
        d is the dimension of the restricted subspace
        """
        x_large = super().get_object_parameters()
        size = len(self.founder.positions.flatten())
        x_large.reshape((-1, size))
        indices = self.atoms_within_cutoff
        size = (self.cmaes.pop_size, self.cmaes.dimension)
        x = np.empty(size)
        for j, index in enumerate(indices):
            x[:, 3*j:3*j+3] = x_large[:, 3*index:3*index + 3]
        return x

    def _mutate_individual(self, individual):
        indices = self.atoms_within_cutoff
        ext_x_ind = individual.positions.flatten()
        x_ind = self.cmaes._mutate_single()
        for j, index in enumerate(indices):
            ext_x_ind[3 * index:3 * index + 3] = x_ind[3 * j:3 * j + 3]
        individual.positions = ext_x_ind.reshape(-1, 3)

    def _prepare_distribution_parameters(self, indices):
        dim = len(indices)*3
        mean = self.founder.positions.ravel()
        matrix = self._calculate_rank_s_matrix()*np.square(self._cr)
        C_0 = self._covariance
        C_0_r = np.zeros((dim, dim))
        matrix = np.diag(matrix)
        d_matrix = np.zeros(dim)
        d_mean = np.zeros(dim)
        for i, ind in enumerate(indices):
            d_matrix[3*i:3*i + 3] = matrix[3*ind:3*ind + 3]
            d_mean[3*i:3*i + 3] = mean[3*ind:3*ind + 3]
            for j, indj in enumerate(indices):
                C_0_r[3*i:3*i + 3, 3*j:3*j + 3] = C_0[3*ind:3*ind + 3,
                                                      3*indj:3*indj + 3]
        matrix = np.diag(d_matrix)
        self._cmaes_params['mean'] = d_mean
        self._cmaes_params['covariance'] = C_0_r + matrix

    def _get_fitness_calculator(self):
        return self.fitness_calc_obj(
            self.population, self.atoms_within_cutoff)

    def _prepare_strategy_parameters(self, indices):
        dim = len(indices)*3
        sp = self._strategyparameters
        pop_size = sp.pop_size
        weights = sp.weights
        c_sigma = sp.c_sigma
        d_sigma = sp.d_sigma
        c_c = sp.c_c
        c_1 = sp.c_1
        c_mu = sp.c_mu
        alpha_cov = sp.alpha_cov
        std_min = sp.std_min
        c_g = sp.c_g
        if hasattr(self, '_use_reduced_pop_size'):
            if self.use_reduced_population_size is True:
                pop_size = None
                weights = None
                c_sigma = None
                d_sigma = None
                c_c = None
                c_mu = None
        strategy_params = StrategyParameters(
            dim, pop_size=pop_size,
            weights=weights, c_sigma=c_sigma, d_sigma=d_sigma,
            c_c=c_c, c_1=c_1, c_mu=c_mu, alpha_cov=alpha_cov,
            std_min=std_min, c_g=c_g)
        self.set_strategy_parameters(strategy_params)

    def _set_up_cmaes(self):
        if not self._external_cmaes:
            indices = self.atoms_within_cutoff
            self._prepare_distribution_parameters(indices)
            self._prepare_strategy_parameters(indices)
            self._cmaes = self.cmaes_obj(**self._cmaes_params)


class RSPopulationEvolverGrad(RSPopulationEvolver, GpPopulationEvolver):
    """
    Parameters
    ----------
    c_alpha : float
        coefficient describing the relevance of the gradient term

    nn_cutoff : float
        cutoff radius, in Angstrom, including the atoms used to build the
        rank-s matrix

    c_r : float
        coefficient to control the contribution of the rank-s matrix to
        the initial covariance matrix

    founder : evpd.core.individual instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    kwargs : other keyword arguments necessary to initialize a ``PopulationEvolver``
        instance


    Notes
    -----
    Similar to ``RSPopulationEvolver`` but gradients are also used.
    are moved.
    """

    cmaes_obj = GpCMAES
    fitness_calc_obj = RSFitnessGradientCalculator

    def __init__(self, c_alpha, nn_cutoff, c_r, founder, **kwargs):
        GpPopulationEvolver.__init__(self, c_alpha=c_alpha,
                                     nn_cutoff=nn_cutoff, c_r=c_r,
                                     founder=founder, **kwargs)

    def _set_up_cmaes(self):
        if not self._external_cmaes:
            indices = self.atoms_within_cutoff
            self._prepare_distribution_parameters(indices)
            self._prepare_strategy_parameters(indices)
            self._cmaes = self.cmaes_obj(**self._cmaes_params)
            self.cmaes.set_gradient_coefficient(self._c_alpha)
        else:
            self._c_alpha = self.cmaes.gradient_coefficient


class RSPopulationEvolverMetamodel(RSPopulationEvolver):
    """ RS Fitness Calculator with a metamodel to be trained on-the-fly

    Parameters
    ----------
    metamodel : a ``Metamodel`` object that will be used to make the energy
        predictions

    dataset : string
        the name of the .hdf5 file which will be used to write/read the
        training data

    nn_cutoff : float
        cutoff radius, in Angstrom, including the atoms used to build the
        rank-s matrix

    c_r : float
        coefficient to control the contribution of the rank-s matrix to
        the initial covariance matrix

    founder : evpd.core.individual instance
        an ``Individual`` object representing the initial individual.
        The mean of the population is taken as the atomic position of this
        individual. It should ideally be an atomic configuration not too
        far from the global minimum in the PES.
        The founder must have a calculator set.

    min_generation : int. Default 0
        use the meta-model only in the current generation is
        larger or equal than ``min_generation``

    train_kwargs : dict
        the keyword-argument values used to train the metamodel

    kwargs : other keyword arguments necessary to initialize a ``PopulationEvolver``
        instance
    """

    fitness_calc_obj = MetaRSFitnessCalculator

    def __init__(self, metamodel, dataset, nn_cutoff, c_r, founder,
                 min_generation=0, train_kwargs=dict(), **kwargs):
        super().__init__(nn_cutoff=nn_cutoff, c_r=c_r,
                         founder=founder, **kwargs)
        self.metamodel = metamodel
        self.dataset = dataset
        self.min_generation = min_generation
        self.train_kwargs = train_kwargs

    def _get_fitness_calculator(self):
        return self.fitness_calc_obj(self.population,
                                     self.atoms_within_cutoff,
                                     self.metamodel,
                                     self.dataset,
                                     min_generation=self.min_generation,
                                     train_kwargs=self.train_kwargs)


class AnalizeRun:
    """ Helper class for analyzing the evolution of the population

    Parameters
    ----------
    Evolver : ``PopulationEvolver`` derived instance
        the evolver to be analyzed.
        If Evolver is set to None, then the class
        can be used to analyze an already existing
        simulation dataframe (set with the method 
        ``load_dataframe``.
    """
    columns = ['generation', 'seed', 'defect_pos',
               'c_alpha (A^2/eV)', 'nn_dist (A)', 'c_r',
               'step_size', 'mean E (eV)', 'std E (eV)',
               'max E (eV)', 'min E (eV)',
               'grad (eV/A)', 'ffe']

    def __init__(self, evolver):
        if evolver:
            if not isinstance(evolver, PopulationEvolver):
                raise TypeError('Evolver must be a PopulationEvolver '
                                '(sub)instance')
            self._evolver = evolver
            self._ffe = 0
            self._data = []

    @property
    def evolver(self):
        return self._evolver

    def initialize(self):
        """ Initialize the evolution

        Returns
        -------
        dataframe : pandas DataFrame with the elements for
            generation 0
        """
        if MPI.COMM_WORLD.Get_rank() == 0:
            logger.debug('Initializing the Run')
        try:  # if starting from a previous run
            generation = self.evolver.cmaes.g - 1
        except AttributeError:
            generation = 0
        defect_pos = str(self.evolver.founder.defect_position)
        self._defect_pos = defect_pos
        if isinstance(self.evolver, GpPopulationEvolver):
            c_alpha = self.evolver.gradient_coefficient
        else:
            c_alpha = 0
        self._c_alpha = c_alpha
        if isinstance(self.evolver, SSPopulationEvolver):
            nn_dist = self.evolver._nn_cutoff
            c_r = self.evolver._cr
        else:
            nn_dist = np.nan
            c_r = 0
        self._nn_dist = nn_dist
        self._c_r = c_r
        sigma = self.evolver._step_size
        energy = self.evolver.founder.cost
        self._ffe += self.evolver.founder.total_energy_calculations
        ffe = self._ffe
        if isinstance(self.evolver, GpPopulationEvolver):
            grad = self.evolver.founder.calc.get_forces().ravel()
            grad = np.linalg.norm(grad)
        else:
            grad = np.nan

        row = {'generation': generation,
               'seed': self.evolver._random_seed,
               'defect_pos': defect_pos, 'c_alpha (A^2/eV)': c_alpha,
               'nn_dist (A)': nn_dist, 'c_r': c_r, 'step_size': sigma,
               'mean E (eV)': energy, 'std E (eV)': 0,
               'max E (eV)': energy, 'min E (eV)': energy,
               'grad (eV/A)': grad, 'ffe': ffe}
        self._data.append(row)
        return pd.DataFrame(self._data, columns=self.columns)

    def evolve(self):
        """ Evolve generation-by-generation

        Yields
        ------
        dataframe : pandas DataFrame updated to the
            current generation
        """
        for params in self.evolver.evolve_population():
            energies = -params['fitnesses']
            generation = params['g']
            C = params['C']
            index = np.argsort(-energies)
            emin = np.min(energies)
            emax = np.max(energies)
            emean = np.mean(energies)
            estd = np.std(energies)
            ffe = self._ffe
            for ind in self.evolver.population:
                ffe += ind.total_energy_calculations
            self._ffe = ffe
            if isinstance(self.evolver, GpPopulationEvolver):
                grad = self.evolver.cmaes.gradients
                grad = np.linalg.norm(np.mean(grad, axis=0))
                c_alpha = self.evolver.cmaes.gradient_coefficient
            else:
                grad = np.nan
                c_alpha = 0
            sigma = params['step_size']
            row = {'generation': generation,
                   'seed': self.evolver.cmaes.random_seed,
                   'defect_pos': self._defect_pos,
                   'c_alpha (A^2/eV)': c_alpha,
                   'nn_dist (A)': self._nn_dist, 'c_r': self._c_r,
                   'step_size': sigma, 'mean E (eV)': emean,
                   'std E (eV)': estd,
                   'max E (eV)': emax, 'min E (eV)': emin,
                   'grad (eV/A)': grad, 'ffe': ffe}

            self._data.append(row)
            dataframe = pd.DataFrame(self._data, columns=self.columns)
            self.dataframe = dataframe.copy()
            yield dataframe

    def run(self):
        """ Evolve the population until a termination criterion is met.

        Returns
        -------
        dataframe : pandas DataFrame
            stores information about the evolution of ``Evolver``
        """
        data_init = self.initialize()
        yield data_init
        for dataframe in self.evolve():
            yield dataframe

    @staticmethod
    def plot_data_vs_generation(df, keys, other_keys=None, samples=[1],
                                serrors=[0], alpha=0.05, **kwargs):
        """ Plot the evolution of ``key`` and eventually ``other_keys``,
        in ``df`` with respect
        to the generation number.

        Parameters
        ----------
        df : pandas data frame
            it must contains at least 2 columns:
            'generation', with the generation number,
            and ``key``.

        keys : list of strings
            the column labels in ``df`` to be plotted.
            If this is an averaged value, the sample stds
            are given by ``se``

        other_keys : None or list of strings
            the eventual other column labels to be plotted

        samples : list of int
            if one of ``keys`` is an average, it is its sample size.

        serrors : list of float
            if one of ``keys`` is an average, it is its sample std.
            This will be used to calculate the confidence intervals

        alpha : float in (0, 1)
            defines the wished (1-alpha)*100% confidence interval

        kwargs : dictionary
            keyword-value pairs for tuning the plot parameters
            see documentation of ``pandas.DataFrame.plot.line``
        """
        figsize = kwargs.get('figsize', (10, 8))
        use_index = False
        fontsize = kwargs.get('fontsize', 24)
        linewidth = kwargs.get('linewidth', 2)
        x = df['generation'].values
        cfs = []
        cfas = []
        cfbs = []
        for i, key in enumerate(keys):
            s = samples[i]
            se = serrors[i]
            cf = scipy.stats.t.ppf((2-alpha)/2, s-1)
            cfa = df[key] - cf*se/np.sqrt(s)
            cfb = df[key] + cf*se/np.sqrt(s)
            cfs.append(cf)
            cfas.append(cfa)
            cfbs.append(cfb)
        if len(keys) == 1:
            ax = df.plot.line('generation', keys, figsize=figsize,
                              use_index=use_index, fontsize=fontsize,
                              linewidth=linewidth,
                              color='black')
        else:
            ax = df.plot.line('generation', keys, figsize=figsize,
                              use_index=use_index, fontsize=fontsize,
                              linewidth=linewidth)
        ax.plot(x, df[other_keys].values, linestyle='--', color='blue')
        for i, (cfa, cfb) in enumerate(zip(cfas, cfbs)):
            ax.plot(x, cfa.values, color='red', linestyle='--', linewidth=2)
            ax.plot(x, cfb.values, color='red', linestyle='--', linewidth=2)
            if i == 0:
                ax.fill_between(x, cfa.values, cfb.values, alpha=0.5, color='red',
                                label='{:d}% C.I.'.format(int((1 - alpha)*100)))
            else:
                ax.fill_between(x, cfa.values, cfb.values, alpha=0.5, color='red')
        plt.legend(loc=1, prop={'size': 18})
        return ax

    def plot_energy_vs_generation(self, **kwargs):
        """ Plots the evolution of the mean population energy
        as a function of the generation.

        Parameters
        ----------
        kwargs : dictionary
            keyword-value pairs for tuning the plot parameters
            see documentation of ``pandas.DataFrame.plot.line``
        """
        df = self.dataframe
        keys = ['max E (eV)', 'min E (eV)']
        s = self.evolver.cmaes.pop_size
        se = df['std E (eV)']*np.sqrt(s/(s-1))
        AnalizeRun.plot_data_vs_generation(df, ['mean E (eV)'], keys,
                                           samples=[s], serrors=[se],
                                           **kwargs)
        plt.show()

    def load_dataframe(self, df):
        """ Use this method to analyze a proper simulation dataframe
        without need of running the whole evolutionary process

        Parameters
        ----------
        df : pandas DataFrame
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError('The dataframe should be a pandas one')
        if not set(self.columns).issubset(set(df.columns)):
            raise ValueError('The dataframe should possess at least '
                             'these columns: {}'.format(self.columns))
        self.dataframe = df.copy()



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import copy
import numpy as np
import os

import h5py
from mpi4py import MPI

from dscribe.descriptors import (SOAP, MBTR, LMBTR, ACSF, SineMatrix,
    EwaldSumMatrix)

from clinamen.utils import get_structure_id
from clinamen.descriptors.finger_descriptors_cython import StructureFingerprint
from clinamen.descriptors.utils import write_descriptors, read_descriptors_by_id
from clinamen.evpd import ROOT_LOGGER as logger


COMM_ = MPI.COMM_WORLD
PROC_RANK = COMM_.Get_rank()


class DescriptorsGenerator:
    """ Base class that will produce the desired material descriptors.

    It is expected that one will set the structures for which the descriptors
    will be calculated with :attr:`structures` and then the descriptors for
    these structures can be calculated with :meth:`create_descriptors`.
    """
    available_descriptors = [
        'fingerprints', 'soap', 'mbtr', 'lmbtr', 'acsf', 'sinematrix',
        'esm'
    ]

    descriptors_map = {
        'fingerprints': StructureFingerprint,
        'soap': SOAP,
        'mbtr': MBTR,
        'lmbtr': LMBTR,
        'acsf': ACSF,
        'sinematrix': SineMatrix,
        'esm': EwaldSumMatrix
    }

    def __init__(self, descriptor_name, descriptor_kwargs,
                 creation_kwargs=None):
        """
        Initialize the object that will be used in order to generate the
        descriptors.

        Parameters:
        -----------
        descriptor_name : string
            string that defines the type of descriptors to be used.
            Must be either: 'fingerprints', 'soap', 'mbrt', 'lmbtr', 'acsf',
            'sinematrix', 'esm'

        descriptor_kwargs : dictionary
            keyword-value pairs needed to initialize the descriptor object.
            Each descriptor has its own arguments, so one should check the
            documentation for each descriptor.

        creation_kwargs : dictionary
            optional keyword-value pairs to specify how the descriptors should
            be created. These are the optional parameters used by the
            ``create`` method of a DScribe descriptor object
        """
        self._descriptor_name = descriptor_name
        self._descriptor_kwargs = descriptor_kwargs
        if creation_kwargs is None:
            self._creation_kwargs = dict()
        else:
            self._creation_kwargs = creation_kwargs

        if self._descriptor_name.lower() not in self.available_descriptors:
            raise ValueError(f"Descriptor '{self._descriptor_name}' "
                             "is not available")
        self._descriptor_object = self.descriptors_map[
            self._descriptor_name.lower()]

        # Fingerprint descriptors have different interface
        self._using_fingerprints = (
            True if self.descriptor_name.lower() == 'fingerprints' else False
        )

        self._structures = None

    def _create_descriptors(self, descriptor_creator, atoms):
        if not self._using_fingerprints:
            descriptors = descriptor_creator.create(atoms,
                                                    **self._creation_kwargs)
        else:
            descriptors = descriptor_creator.get_descriptors()
        return descriptors

    def create_descriptors(self):
        """ Create the descriptors for each structure in
        :attr:`self.structures`
        """
        if self.structures is None:
            raise RuntimeError('No structures for which descriptors have to'
                               'be calculated have been given')
        if not self._using_fingerprints:
            init_descriptor = self._descriptor_object(
                **self._descriptor_kwargs)
            for atoms in self.structures:
                yield self._create_descriptors(init_descriptor, atoms)
        else:
            for atoms in self.structures:
                init_descriptor = StructureFingerprint(
                    atoms, **self._descriptor_kwargs)
                yield self._create_descriptors(init_descriptor, atoms)

    @property
    def structures(self):
        """ list of atomic structures for which we want to calculate the
            descriptors
        """
        return self._structures

    @structures.setter
    def structures(self, structures):
        self._structures = structures

    @property
    def descriptor_name(self):
        return self._descriptor_name

    @property
    def descriptor_properties(self):
        return copy.deepcopy(self._descriptor_kwargs)


class DescriptorsDatabase:
    """ Class to manage calculated descriptors """
    def __init__(self, hdf5_filename, initialized_descriptors_generator,
                 tag='descriptors', batch_size=128):
        """
        Parameters:
        -----------
        hdf5_filename : string
            the name of the file where the descriptors have to be read,
            written or appended.
            If a file with name ``hdf5_filename`` already exists, new
            descriptors will be appended.

        initialized_descriptors_generator: :class:`DescriptorsGenerator`
            instance
            The instance used to generate descriptors for the dataset

        tag : string
            the name of the dataset (``dataset.attrs['name']``)

        batch_size : int, default 128
            the number of descriptors to read/write at one time.
            If None, all descriptors will be read/written at once
        """
        self._filename = hdf5_filename

        self.descriptors_generator = initialized_descriptors_generator
        self.tag = tag
        self.batch_size = batch_size

    @property
    def database_exists(self):
        if os.path.isfile(self.filename):
            self._database_exists = True
        else:
            self._database_exists = False
        return self._database_exists
        
    @property
    def filename(self):
        return self._filename
 
    def get_structure_id(self, atoms):
        """ Get the id of an atomic structure

        Parameters:
        -----------
        atoms : ``ase.atoms.Atoms`` instance
            an atomic structure
        """
        positions = atoms.positions.ravel()
        atomic_numbers = atoms.get_atomic_numbers()
        cell = atoms.cell[:].ravel()
        pbc = atoms.pbc
        id_ = get_structure_id(positions, atomic_numbers, cell, pbc)
        return id_

    @property
    def existing_ids(self):
        """ Return a tuple with existing structure ids in the database """
        return self._get_existing_ids()

    def _get_existing_ids(self):
        if not self.database_exists:
            if PROC_RANK == 0:
                logger.debug('No database has been found.')
            return tuple()
        else:
            visited_ids = []
            if PROC_RANK == 0:
                with h5py.File(self.filename, 'r') as f:
                    for val in f['X'].keys():
                        visited_ids.append(val)
            visited_ids = COMM_.bcast(visited_ids, root=0)
            COMM_.Barrier()
            return tuple(visited_ids)

    def write_descriptors(self, structures):
        """ Write a list of atomic structures to the database

        Parameters:
        -----------
        structures: list
            a list of ``ase.atoms.Atoms`` instances of which
            we want to write the descriptors on the database
        """
        indices = np.arange(len(structures))
        no_chunks = int(np.ceil(len(indices)/self.batch_size))
        c_indices = [self.batch_size*i for i in range(1, no_chunks)]
        chunks = np.array_split(indices, c_indices)
        for chunk_no, chunk in enumerate(chunks):
            existing_ids = list(self.existing_ids)
            if PROC_RANK == 0:
                logger.debug(f'Found {len(existing_ids)} existing ids.')
            existing_indices = []
            ids = []
            for indx in chunk:
                id_ = self.get_structure_id(structures[indx])
                if id_ in existing_ids:
                    existing_indices.append(indx)
                else:
                    ids.append(id_)
                    existing_ids.append(id_)
            if PROC_RANK == 0:
                logger.debug(f'{len(ids)} ids to process.')
            ch_structures = [structures[i] for i in chunk
                             if i not in existing_indices]
            if len(ch_structures) != len(ids):
                raise RuntimeError('Length of structures in the batch does '
                                   'not equal the length of the ids')
            self.descriptors_generator.structures = ch_structures
            if PROC_RANK == 0:
                logger.debug(f'Processing chunk {chunk_no + 1}/{len(chunks)}. '
                             f'{len(ids)} descriptors for '
                             f'database: {self.filename}')
                descrs = self.descriptors_generator.create_descriptors()
                array_descrs = []
                for desc in descrs:
                    array_descrs.append(desc.ravel())
                if len(array_descrs) > 0:
                    logger.debug(f'Writing {len(ids)} descriptors to '
                                 f'database: {self.filename}.'
                                 f'{len(existing_indices)} structures were '
                                 'already processed.')
                    array_descrs = np.vstack(array_descrs)
                    write_descriptors(self.filename,
                                      array_descrs, None, ids, self.tag)
                    logger.debug(f'{len(ids)} descriptors have been '
                                 f'written to : {self.filename}')
                else:
                    logger.debug(f'No descriptors to add to '
                                 f'database: {self.filename}')

    def read_descriptors(self, ids=None):
        """ Read the descriptors from the database

        Parameters:
        -----------
        ids : list or None. Default None
            if a list, is a list of identifiers for the structures whose
            descriptors we want to retrieve. If None, all descriptors in
            the database are returned.
        """
        if ids is None:
            ids = self.existing_ids
        else:
            ids = tuple(ids)
        descriptors, jacobians, indices = read_descriptors_by_id(self.filename,
                                                                 ids)
        correct_indices = np.arange(len(descriptors))
        if (correct_indices != indices).any():
            raise RuntimeError('Error: for some reason, descriptors were not '
                               f'read correctly from {self.filename}')
        descriptors = np.vstack(descriptors)
        if jacobians is None:
            return descriptors, None
        else:
            jacobians = np.vstack(jacobians)
            return descriptors, jacobians



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import json
import numpy as np
import hashlib
import h5py

from ase.io.vasp import write_vasp
from ase.atoms import Atoms
from ase.io.trajectory import Trajectory

from clinamen import DESCRIPTORS_FLOAT_DTYPE as DTYPE


def counter(function):
    function._callcount = 0
    def decorated(*args, **kwargs):
        function._callcount += 1
        decorated.count = function._callcount
        return function(*args, **kwargs)
    return decorated


class GeneralizedEncoder(json.JSONEncoder):
    """ Simple encoder class that allows np.ndarrays
    encoding by using the array tolist() method
    """
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        return json.JSONEncoder.default(self, o)


def get_structure_id(positions, atomic_numbers, cell, pbc):
    """ Obtain an (quasi)unique identifier given an atomic structure

    Parameters
    ----------
    positions: 1D array of shape (n_atoms*3, )
        the flattened atomic positions

    atomic_numbers: 1D array of shape (n_atoms, )
        the atomic numbers identifying the atoms in the system

    cell: 1D numpy array of shape (9, )
        the flattened unit cell vectors

    pbc: 1D array of bools of shape (3, )
       each entry determines if pbc are applied to that direction

    Returns
    -------
    hashed : string
        the string digest of the hash function associated to this
        atomic structure
    """
    no_dec = 10
    dt = '<f8'
    m = hashlib.sha256()
    positions = np.array(positions, dtype=dt).round(no_dec)
    atomic_numbers = np.array(atomic_numbers, dtype='<u1')
    cell = np.array(cell, dtype=dt).round(no_dec)
    pbc = np.array(pbc, dtype='<u1')
    if positions.ndim != 1:
        raise ValueError('Atomic positions must be given as a 1D array')
    if atomic_numbers.ndim != 1:
        raise ValueError('Atomic numbers must be given as a 1D array')
    if cell.ndim != 1 or len(cell) != 9:
        raise ValueError('Unit cell must be given as a 1D array '
                         'with 9 entries')
    if pbc.ndim != 1 or len(pbc) != 3:
        raise ValueError('PBC must be given as a 1D array with 3 entries')
    no_atoms = len(atomic_numbers)
    if not len(positions) == 3*no_atoms:
        raise ValueError(f'The structure has {no_atoms} atoms, but '
                         'the atomic positions describe '
                         f'{len(positions)//3} atoms')

    structure_id = [positions, atomic_numbers, cell, pbc]
    structure_id = np.concatenate(structure_id)

    m.update(structure_id.tostring(order='C'))
    hashed = m.hexdigest()

    return hashed


def get_id(atoms):
    """ Return the id of the structure.

    Parameters
    ----------
    atoms : ase.Atoms instance
        the structure whose id has to be calculated

    Returns
    -------
    id_ : string
        the id of the structure
    """
    positions = atoms.positions.ravel()
    atomic_numbers = atoms.get_atomic_numbers()
    cell = atoms.cell[:].ravel()
    pbc = atoms.pbc
    id_ = get_structure_id(positions, atomic_numbers, cell, pbc)
    return id_


def make_atoms(cell, numbers, positions, pbc):
    """ Return an ase.Atoms instance from structural data

    Parameters
    ----------
    cell : 3x3 array
        the Cartesian coordinates of the structure cell

    numbers : 1D array of length N
        the atomic numbers of the atoms in the structure

    positions : 2D array of shape (N, 3)
        the Cartesian positions of the atoms in the system

    pbc : 1D array of 3 bools
        whether pbc are applied along each of the 3 dimensions

    Returns
    -------
    atoms : ase.Atoms instance
        the Atoms object representing the structure
    """
    atoms = Atoms(cell=cell, numbers=numbers, positions=positions,
                  pbc=pbc)
    return atoms


def generate_fingerprints(traj_filename, descr_kwargs, verbose=True):
    """ For each structure in a .traj file calculates the structure
    fingerprints descriptors according to ``descr_kwargs``

    Parameters
    ----------
    traj_filename : string
        the path where the .traj file is located

    descr_kwargs: dict
        the key:value pairs needed to initialize a StructureFingerprint
        instance

    verbose : bool. Default True
        if True, the function will print the progress of the calculation

    Returns
    -------
    descriptor_name : string
        the name of the hdf5 file where the descriptors are written.

    Notes
    -----
    The function will write the text file:
    visited_ids.txt : List of structure ids already visited.
    If the function is interrupted, it will read this file to see
    for which structures features were already calculated
    """
    descriptor_name = [str(descr_kwargs[x]) for x in descr_kwargs.keys()]
    descriptor_name = '_'.join(descriptor_name)
    ids = []
    feats = []
    traj_file = Trajectory(traj_filename)
    tot = len(traj_file)
    unique_structs = 0
    if os.path.isfile('visited_ids.txt'):
        visited_ids = []
        with open('visited_ids.txt', 'r') as f:
            for line in f:
                visited_ids.append(line.strip())
    else:
        visited_ids = []
    for i, atoms in enumerate(traj_file):
        if verbose:
            print(f'Structure {i}/{tot}')
        id_ = get_id(atoms)
        if id_ in visited_ids:
            continue
        if verbose:
            print('id: ', id_)
            print(len(atoms))
        if id_ in ids:
            if verbose:
                print('The structure has already been visited')
            xx = np.where(np.array(ids) == id_)[0]
            if verbose:
                print(f'Is the structure of system {xx}')
        else:
            unique_structs += 1
            feat = StructureFingerprint(atoms, **descr_kwargs).get_descriptors()
            if verbose:
                print(f'Number of features: {len(feat)}')
            ids.append(id_)
            feat = np.atleast_2d(feat)
            write_descriptors('finger_descriptors' + descriptor_name + '.hdf5',
                              feat, None, [id_], 'clustering_set')
            with open('visited_ids.txt', 'a') as f:
                f.write(id_ + '\n')
            if verbose:
                print('Written')
    return descriptor_name


def generate_evolution_trajectory(evolution_history, traj_filename):
    """ From a evolution history hdf5 file, generate the evolution
    trajectory. Note that the individuals are sorted by generation.
    Starting with the Founder

    Parameters
    ----------
    evolution_history : string
        the path to the evolution_history.hdf5 file

    traj_filename : string
        the name of the .traj file to be generated    

    Notes
    -----
    The function will write the text file:
    mapping_ids_structure_names.txt
    The first column has the structure name, as read from
    ``evolution_history`` and the second column the corresponding structure id
    """
    if os.path.isfile(traj_filename):
        print(f'File {traj_filename} already exists')
    else:
        print(f'Creating {traj_filename}')
        structs = []
        structs_names = []
        structs_ids = []
        with h5py.File(evolution_history, 'r') as f:
            cell = np.array(f['system']['cell'])
            numbers = np.array(f['system']['atomic_numbers'])
            pbc = np.array(f['system']['pbc']).astype(bool)
            no_atoms = len(numbers)
            all_keys = []
            for key in f.keys():
                if key != 'system':
                    all_keys.append(key)
            gen_indices = [int(x.split('_')[1]) for x in all_keys]
            sorting_ind = np.argsort(gen_indices)
            all_keys = np.array(all_keys)[sorting_ind]
            for key in all_keys:
                gen_positions = np.array(f[key]['positions'])
                for positions, label in zip(gen_positions,
                                            np.array(f[key]['labels'])):
                    positions = positions.reshape((no_atoms, 3))
                    atoms = make_atoms(cell, numbers, positions, pbc)
                    structs.append(atoms)
                    structs_names.append(label.decode())
                    structs_ids.append(get_id(atoms))

        all_traj = Trajectory(traj_filename, 'a')
        for struct in structs:
            all_traj.write(struct)
        all_traj.close()

        print('Creating mapping file')
        mapping = open('mapping_id_structure_names.txt', 'w')
        for i_name, id_ in zip(structs_names, structs_ids):
            mapping.write(i_name + '\t' + id_ + '\n')
        mapping.close()


def retrieve_structures_names_and_ids_from_file(mapping_file):
    """ Read a mapping file and retrieve the pairs of strucure ids and
    structure names.

    Parameters
    ----------
    mapping_file : string
        The path to a text file containing structure names on the first
        column and corresponding structure ids on the second one

    Returns
    -------
    structures, ids : tuple
		the first is a list of structure names and the second a list of
		corresponding ids
	    """
    structures = []
    ids = []
    with open(mapping_file, 'r') as f:
        for line in f:
            struct, id_ = line.strip().split()
            structures.append(struct) 
            ids.append(id_)
    return structures, ids


def get_descriptors(file_name, mapping_file):
    """ Read a hdf5 file containing the descriptors for some structures.

    Parameters
    ----------
    file_name : string
        path to the hdf5 file containing the calculated descriptors

    mapping_file : string
        The path to a text file containing structure names on the first
        column and corresponding structure ids on the second one

    Returns
    -------
    descriptors, ordered_structures : tuple
        The first element is a 2D array with the descriptors. Each row
        represents a structure. The second file is a 1D array with the
        corresponding structure names.
    """
    structures, ids = retrieve_structures_names_and_ids_from_file(mapping_file)
    descriptors = []
    ordered_structures = []
    structures = np.array(structures)
    with h5py.File(file_name, 'r') as f:
        dataX = f['X']
        for id_ in dataX.keys():
            index = np.where(np.array(ids) == id_)[0]
            structure = structures[index]
            assert len(structure) == 1
            ordered_structures.append(structure[0])
            descriptors.append(np.array(dataX[id_], dtype=DTYPE))
        print(f'Found {len(descriptors)} structures')

    descriptors = np.vstack(descriptors)
    
    return descriptors, np.array(ordered_structures)



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy

ext_modules = [
    Extension('finger_descriptors_cython',
              sources=['finger_descriptors_cython.pyx', 'finger_descriptors_c.c',
                       'basic_operations.c'],
              libraries=['m'],
              include_dirs=[numpy.get_include(), '.']
              )
]

setup(name='descriptors',
      ext_modules=cythonize(
          ext_modules, 
          compiler_directives={'language_level': sys.version_info[0]}))



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from collections.abc import MutableSequence
import numpy as np
import pandas as pd

from clinamen.evpd import ROOT_LOGGER as logger
from clinamen.evpd.core.individual import Individual


class BadPopulationMember(Exception):
    """ Raise the exception when one tries to add to a ``Population``
    instance an object which is not an ``Individual`` instance
    """
    pass


class Population(MutableSequence):
    """ A population is a group of individual of a given size.

    Parameters
    ----------
    individuals : a list or tuple of individuals.
        the container of individuals forming the population.
        It can also be a single individual or empty.
    """
    def __init__(self, *individuals): 
        self._individuals = []
        # passing variable number of arguments
        if hasattr(individuals, '__iter__') and len(individuals) > 1:
            for ind in individuals:
                self._check_type(ind)
                self._individuals.append(ind)
        # empty
        elif hasattr(individuals, '__iter__') and len(individuals) == 0:
            pass
        # passing a list or tuple
        elif ((isinstance(individuals[0], tuple) 
              or isinstance(individuals[0], list)) and len(individuals) == 1):
            for ind in individuals[0]:
                self._check_type(ind)
                self._individuals.append(ind)
        # single element
        elif len(individuals) == 1:
            self._check_type(individuals[0])
            self._individuals.append(individuals[0])
        else:
            self._check_type(individuals)
            self._individuals.append(individuals)
        
    def _check_type(self, obj):
        if not isinstance(obj, Individual):
            raise BadPopulationMember('Only instances of type Individual can '
                                      'be part of a Population instance')

    def __len__(self):
        return len(self._individuals)
 
    def __getitem__(self, key):
        """ Return another Population instance if key is a slice.
        Note that the individuals are not copied, similar behavior of a
        ``list``"""
        if isinstance(key, slice):
            return self.__class__(self._individuals[key])
        else:
            return self._individuals[key]

    def __setitem__(self, key, value):
        isiter = hasattr(value, '__iter__')
        if isinstance(key, int) and isinstance(value, Individual):
            self._individuals[key] = value
            logger.info('Individual number "{}" has been replaced'.format(key))
        elif isinstance(key, slice) and isiter:
            for i, individual in enumerate(value):
                self._check_type(individual)
                step = key.step if key.step else 1
                self._individuals[key.start + i * step] = individual 
            logger.info('Individuals "{}" have been '
                        'replaced'.format(':'.join([str(key.start),
                                                    str(key.stop),
                                                    str(key.step)])))
        else:
            self._check_type(value)
 
    def __delitem__(self, key):
        if isinstance(key, (int, slice)):
            del self._individuals[key]
        else:
            raise TypeError('Indices must be integers or slices')

    def insert(self, index, value):
        isseq = hasattr(value, '__len__') and hasattr(value, '__getitem__')
        if isseq and isinstance(value, Individual): # Individuals are iterables
            pass
        elif isseq: # another iterable
            for individual in value:
                self._check_type(individual)
        else:
            self._check_type(value)
        if isinstance(index, int):
            self._individuals.insert(index, value)
        else:
            raise TypeError('The index must be an integer')

    @property
    def individuals_fitness(self):
        """ Return a list with the fitness of each individual """
        self._individuals_fitness = [x.fitness for x in self]
        return self._individuals_fitness

    @property
    def dataframe(self):
        data = [x.data for x in self]
        df = pd.DataFrame(data=data)
        df.rename_axis('Name', inplace=True)
        df.reset_index(inplace=True)
        df.index = df.index.map(int)
        return df
        
    def sort(self):
        fitnesses = self.individuals_fitness
        indices = np.argsort(fitnesses)
        old_ind = self._individuals[:]
        for i, index in enumerate(indices):
           self._individuals[i] = old_ind[index]

    def get_equal_individuals(self):
        keys = [x.my_name for x in self]
        equals = {}
        for i, ind in enumerate(self):
            if ind.my_name in keys:
                equals[ind.my_name] = []
                keys.remove(ind.my_name)
                for ind2 in self[i+1:-1]:
                    if ind == ind2:
                        equals[ind.my_name].append(ind2.my_name)
                        keys.remove(ind2.my_name)
        return equals



# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
import os
import h5py
from mpi4py import MPI

from ase import Atoms

from clinamen.evpd import ROOT_LOGGER as logger
from clinamen.evpd.core.population import Population
from clinamen.evpd.core.individual import Individual
from clinamen import INDIVIDUALS_FLOAT_DTYPE as DTYPE


proc_rank = MPI.COMM_WORLD.Get_rank()


def write_train_hdf5(file_name, population, energies, forces, name=None):
    """ Append new data to an existing dataset, if the dataset
    does not exist, create a new one

    Parameters
    ----------
    file_name : string
        the dataset name

    population : `Population` instance
        the new individuals to be added to the dataset

    energies : array-like of shape (n_individuals, )
        the energies of the individuals in ``population``

    forces : 2D array-like of shape (n_individuals, 3*no_atoms)
        the forces of the individuals in ``population``

    name : string. Default None
        the system name. A tag that specifies the system when the
        dataset is created. If the dataset already exists, it checks
        the it corresponds to system ``name``
    """
    comp_kwargs = {'compression': 'gzip', 'compression_opts': 9}
    founder = population[0]
    d = population[0].positions.ravel().shape[0]
    shape_x = (None, d)
    shape_y = (None, )
    shape_dy = (None, d)
    xs = [x.positions.ravel() for x in population]
    X = np.vstack(xs)
    y = np.array(energies)
    dy = np.array(forces)
    y = np.atleast_1d(y)
    X, dy = np.atleast_2d(X, dy)
    new_dataset = True
    if os.path.isfile(file_name):
        new_dataset = False
        # get data to check dataset consistency
        if proc_rank == 0:
            logger.debug('Found an existing dataset with '
                         f'training data: {file_name}')
        with h5py.File(file_name, 'r') as f:
            sys_name = f['system'].attrs['name'].decode()
            atomic_numbers = np.array(f['system']['atomic_numbers'])
            cell = np.array(f['system']['cell'])
            pbc = np.array(f['system']['pbc'])
        if name is not None:
            c0 = (name == sys_name)
        else:
            c0 = True
        for individual in population:
            c1 = np.allclose(atomic_numbers, individual.get_atomic_numbers(),
                             rtol=0, atol=1e-9)
            c2 = np.allclose(cell, individual.get_cell()[:],
                             rtol=0, atol=1e-9)
            c3 = np.allclose(pbc, individual.get_pbc(),
                             rtol=0, atol=1e-9)
            if not np.all([c0, c1, c2, c3]):
                indices = np.where(np.array([c0, c1, c2, c3]) == False)
                error_msg = (f'Individual {individual.my_name} '
                              'does not conform with the dataset '
                             f'{file_name}: ')
                for ind in indices:
                    if ind == 0:
                        ind_msg = f'System name "{name}" is not "{sys_name}". '
                    elif ind == 1:
                        ind_msg = 'Atomic numbers do not agree with those in the dataset. '
                    elif ind == 2:
                        ind_msg = 'Unit cell does not agree with that of the dataset. '
                    elif ind == 3:
                        ind_msg = 'PBC do not agree with those of the dataset. '
                    error_msg += ind_msg
                raise ValueError(error_msg)
    if new_dataset:
        if proc_rank == 0:
            logger.debug('A new dataset with training data will be created '
                         f'with name {file_name}')
            with h5py.File(file_name, 'w') as f:
                system = f.create_group('system')
                if name is None:
                    sys_name = 'system'
                else:
                    sys_name = name
                system.attrs['name'] = np.string_(sys_name)
                atomic_numbers = founder.get_atomic_numbers()
                system.create_dataset('atomic_numbers', data=atomic_numbers,
                                      dtype=np.uint8, **comp_kwargs)
                cell = founder.get_cell()[:]
                system.create_dataset('cell', data=cell, shape=(3, 3),
                                      dtype=DTYPE, **comp_kwargs)
                pbc = founder.get_pbc()
                system.create_dataset('pbc', data=pbc, dtype=bool,
                                      **comp_kwargs)
                data = f.create_group('data')
                data.create_dataset('X', data=X, dtype=DTYPE,
                                    maxshape=shape_x, **comp_kwargs)
                data.create_dataset('y', data=y, dtype=DTYPE,
                                    maxshape=shape_y, **comp_kwargs)
                data.create_dataset('dy', data=dy, dtype=DTYPE,
                                    maxshape=shape_dy, **comp_kwargs)
                logger.debug(f'Added {X.shape} positions, '
                             f'{y.shape} energies and {dy.shape} forces '
                             f'to training set {file_name}')
    else:
        with h5py.File(file_name, 'a') as f:
            dset_X = f['data']['X']
            dset_y = f['data']['y']
            dset_dy = f['data']['dy']
            if proc_rank == 0:
                logger.debug(f'Found {dset_X.shape} positions, '
                             f'{dset_y.shape} energies and {dset_dy.shape} '
                             f'forces in training set {file_name}')
            dset_X.resize(dset_X.shape[0] + X.shape[0], axis=0)
            dset_X[-X.shape[0]:] = X
            dset_y.resize(dset_y.shape[0] + y.shape[0], axis=0)
            dset_y[-y.shape[0]:] = y
            dset_dy.resize(dset_dy.shape[0] + dy.shape[0], axis=0)
            dset_dy[-dy.shape[0]:] = dy
            if proc_rank == 0:
                logger.debug('New data have been appended. There are now: '
                             f'{dset_X.shape} positions, '
                             f'{dset_y.shape} energies and {dset_dy.shape} '
                             f'forces in training set {file_name}')


def read_train_hdf5(file_name):
    proc_rank = MPI.COMM_WORLD.Get_rank()
    with h5py.File(file_name, 'r') as f:
        name = f['system'].attrs['name'].decode()
        atomic_numbers = np.array(f['system']['atomic_numbers'])
        cell = np.array(f['system']['cell'])
        pbc = np.array(f['system']['pbc'])
        data = f['data']
        # Cartesian positions
        X_data = np.array(data['X'])
        # energies
        y_data = np.array(data['y'])
        # forces
        dy_data = np.array(data['dy'])
        X_size = X_data.shape[0]
        if not(X_size == y_data.shape[0] and X_size == dy_data.shape[0]):
            raise ValueError('Dataset is probably corrupt: '
                             f'got {X_size} instances for positions, '
                             f'{y_data.shape[0]} for energies and '
                             f'{dy_data.shape[0]} for forces')
        if proc_rank == 0:
            logger.debug(f'Training dataset {file_name} has been read. '
                         f'Found {X_data.shape} positions, '
                         f'{y_data.shape} energies and {dy_data.shape} '
                         f'forces.')

        return X_data, y_data, dy_data, name, atomic_numbers, cell, pbc


class FitnessCalculator:
    """ Fitness calculator for Population objects

    Parameters
    ----------
    population : Population instance
        the current individual population
    """
    comm = MPI.COMM_WORLD
    comm_size = comm.Get_size()
    proc_rank = comm.Get_rank()

    def __init__(self, population):
        if not isinstance(population, Population):
            raise TypeError("The population must be an instance of "
                            "'evpd.core.population.Population'")
        self._population = population
        self.fitnesses = None
        self.N = len(self.population)
        self.d = len(self.population[0].positions.ravel())

    @property
    def population(self):
        return self._population

    @population.setter
    def population(self, value):
        if not isinstance(value, Population):
            raise TypeError("The population must be an instance of "
                            "'evpd.core.population.Population'")
        self._population = value
        self.N = len(self.population)

    def _check_input_shape(self, input_, shape):
        if not input_.shape == shape:
            raise ValueError("The input has shape {}; expected shape "
                             "{}".format(input_.shape, shape))

    def _update_population_positions(self):
        for i, ind in enumerate(self.population):
            ind.positions = self.x[i].reshape((-1, 3))

    def set_object_parameters(self, x):
        """ Set the object parameters to the individuals in ``self.population``

        Parameters
        ----------
        x : 2D NumPy array
            shape (N, d), with N is the number of individuals in the
            population and d is the dimensionality of the search space
        """
        shape = (self.N, self.d)
        self._check_input_shape(x, shape)
        self.x = x.copy()
        self._update_population_positions()

    def _get_indices_to_process(self, size=None):
        if size is None:
            size = len(self.population)
        batches_size = size//self.comm_size
        batches_rem = size % self.comm_size
        if self.proc_rank < batches_rem:
            ind_start = self.proc_rank*(batches_size + 1)
            ind_end = ind_start + batches_size
        else:
            ind_start = self.proc_rank*batches_size + batches_rem
            ind_end = ind_start + batches_size - 1
        return ind_start, ind_end

    def _gather_fitnesses(self, my_fitnesses):
        fitnesses = []
        if self.proc_rank == 0:
            fitnesses.extend(my_fitnesses)
            for j in range(1, self.comm_size):
                logger.debug(f'Process {self.proc_rank} is receiving the '
                             'fitnesses from slaves with tag 1')
                logger.debug(f'Process {j} is sending its '
                             'fitness to master with tag 1')
                other_fitnesses = self.comm.recv(source=j, tag=1)
                logger.debug(f'Fitness received from processor {j}')
                fitnesses.extend(other_fitnesses)
        else:
            self.comm.send(my_fitnesses, dest=0, tag=1)
        # update also all other processes
        if self.proc_rank == 0:
            logger.debug(f'Starting sync for processes')
        fitnesses = self.comm.bcast(fitnesses, root=0)
        self.comm.Barrier()
        return fitnesses

    def get_fitness(self):
        # update individuals
        start, end = self._get_indices_to_process()
        if self.proc_rank == 0:
            logger.debug(f'The FitnessCalculator in process {self.proc_rank} '
                         f'of {self.comm_size} will calculate the fitness '
                         f'of individuals from {start} to {end}')
            for j in range(1, self.comm_size):
                r_start = self.comm.recv(source=j, tag=32)
                r_end = self.comm.recv(source=j, tag=33)
                logger.debug(f'The FitnessCalculator in process {j} '
                             f'of {self.comm_size} will calculate the fitness '
                             f'of individuals from {r_start} to {r_end}')
        else:
            self.comm.send(start, dest=0, tag=32)
            self.comm.send(end, dest=0, tag=33)
        local_pop = self.population[start:end + 1]
        my_fitnesses = local_pop.individuals_fitness
        self.comm.Barrier()
        fitnesses = self._gather_fitnesses(my_fitnesses)
        self.fitnesses = np.array(fitnesses)
        return np.array(fitnesses)


class FitnessGradientCalculator(FitnessCalculator):
    def _gather_gradients(self, my_grads, start, end):
        grads = np.empty((self.N, self.d), dtype=DTYPE)
        if self.proc_rank != 0:
            data = (start, end)
            self.comm.send(data, dest=0, tag=2)
        if self.proc_rank == 0:
            grads[start:end + 1] = my_grads
            for j in range(1, self.comm_size):
                ostart, oend = self.comm.recv(source=j, tag=2)
                other_grads = np.empty((oend + 1 - ostart, self.d),
                                       dtype=DTYPE)
                self.comm.Recv(other_grads, source=j, tag=3)
                grads[ostart:oend + 1] = other_grads
        else:
            self.comm.Send(my_grads, dest=0, tag=3)
        # update all other processes
        self.comm.Bcast(grads, root=0)
        self.comm.Barrier()
        return grads

    def get_gradients(self):
        """ From a ``Population`` instance with individuals
        with an attached calculator, that RETURNS FORCES, computes
        and returns the gradients for individuals in the population.
        """
        start, end = self._get_indices_to_process()
        if self.proc_rank == 0:
            logger.debug(f'The FitnessCalculator in process {self.proc_rank} '
                         f'of {self.comm_size} will calculate the gradients '
                         f'of individuals from {start} to {end}')
            for j in range(1, self.comm_size):
                r_start = self.comm.recv(source=j, tag=32)
                r_end = self.comm.recv(source=j, tag=33)
                r_rank = self.comm.recv(source=j, tag=34)
                logger.debug(f'The FitnessCalculator in process {r_rank} '
                             f'of {self.comm_size} will calculate the gradients '
                             f'of individuals from {r_start} to {r_end}')
        else:
            self.comm.send(start, dest=0, tag=32)
            self.comm.send(end, dest=0, tag=33)
            self.comm.send(self.proc_rank, dest=0, tag=34)
        my_grads = np.empty(((end - start) + 1, self.d), dtype=DTYPE)
        for i, ind in enumerate(self.population[start:end + 1]):
            my_grads[i] = -ind.get_forces().ravel()
        self.comm.Barrier()
        grads = self._gather_gradients(my_grads, start, end)
        self.gradients = grads.copy()
        return grads


class RSFitnessCalculator(FitnessCalculator):
    """ Fitness calculator for Population objects on a restricted
    subspace.

    Parameters
    ----------
    population : Population instance
        the current individual population

    atoms_within_cutoff : list on integers
        the indices of the atoms within the cutoff forming the
        restricted subspace
    """
    def __init__(self, population, atoms_within_cutoff):
        super().__init__(population)
        self.atoms_within_cutoff = atoms_within_cutoff
        self.founder = self.population[0]

    def set_object_parameters(self, x):
        """ Set the object parameters to the individuals in ``self.population``

        Parameters
        ----------
        x : 2D NumPy array
            shape (N, d), with N is the number of individuals in the
            population and d is the dimensionality of the restricted
            subspace of interest.
        """
        N = self.N
        d = len(self.atoms_within_cutoff)*3
        shape = (N, d)
        self._check_input_shape(x, shape)
        long_x = np.tile(self.founder.positions.ravel()[:, np.newaxis], N).T
        for i, index in enumerate(self.atoms_within_cutoff):
            long_x[:, 3*index:3*index + 3] = x[:, 3*i:3*i + 3]
        self.x = long_x
        self._update_population_positions()


class RSFitnessGradientCalculator(RSFitnessCalculator,
                                  FitnessGradientCalculator):
    def get_gradients(self):
        grads = super().get_gradients()
        indices = self.atoms_within_cutoff
        dim = len(indices)*3
        sub_grads = np.empty((len(grads), dim))
        for i, ind in enumerate(indices):
            sub_grads[:, 3*i:3*i + 3] = grads[:, 3*ind:3*ind + 3]
        self.gradients = sub_grads.copy()
        return sub_grads


class MetaRSFitnessCalculator(RSFitnessCalculator):
    """ Calculator for fitness function using a surrogate
    fitness metamodel. The metamodel is used exclusively
    for predicting the total energy. Forces are saved in
    the training dataset, but are not used for train and
    prediction

    Parameters
    ----------
    population : Population instance
        the current individual population

    atoms_within_cutoff : list on integers
        the indices of the atoms within the cutoff forming the
        restricted subspace

    metamodel : ``MetaModel`` instance
        the fitness surrogate

    data : string
        the name of the file which is used to save/read
        the training data. It will be named ``data``.hdf5

    min_generation : int. Default 0
        use the meta-model only in the current generation is
        larger or equal than ``min_generation``

    train_kwargs : dict
        the keyword-argument pairs to train the metamodel
    """
    def __init__(self, population, atoms_within_cutoff,
                 metamodel, data='Xy.hdf5', min_generation=0,
                 train_kwargs=dict()):
        super().__init__(population, atoms_within_cutoff)
        self.metamodel = metamodel
        self.dataset = data
        self.system_name = None
        self.min_generation = min_generation
        self.current_generation = 0
        self.train_kwargs = train_kwargs
        # keep track of individuals in the last generation
        # which will be used to update the training set
        # and their energy
        self.new_training_population = Population()
        self.new_training_population_indices = []
        self.new_energy_train = []
        self.new_forces_train = []
        # STD on predicted energy for those individuals which will be
        # added to the training set
        self.new_energy_std_train = []

        # local containers for individuals which will be added to training set
        self.new_local_training_population_indices = []
        self.new_local_energy_train = []
        self.new_local_forces_train = []
        # keep track for all generations. These data will be used to
        # train the metamodel
        self.training_population = Population()
        self.energy_train = []
        self.forces_train = []

        # Predicted quantities
        # indices individuals accepted from metamodel
        self.local_valid_indices = []
        self.local_predicted_energies = []
        self.local_predicted_energies_std = []

        # Take initial eventual training data
        self.initialize_training_data()
        # eventually train the model
        self.is_fitted = False
        self.train_model()
        self.reset_counts()  # every time the model is trained, one must
                             # reset these counts

    def reset_counts(self):
        self.new_training_population = Population()
        self.new_training_population_indices = []
        self.new_energy_train = []
        self.new_energy_std_train = []
        self.new_forces_train = []

        self.new_local_training_population_indices = []
        self.new_local_energy_train = []
        self.new_local_forces_train = []

        self.local_valid_indices = []
        self.local_predicted_energies = []
        self.local_predicted_energies_std = []

    def initialize_training_data(self):
        try:
            X_data, y_data, dy_data, name, atomic_numbers, cell, pbc = \
                read_train_hdf5(self.dataset)
        except OSError:
            pass
        else:
            if self.proc_rank == 0:
                logger.info('Found an initial dataset with '
                            f'{len(X_data)} instances.')
            with h5py.File(self.dataset, 'r') as f:
                self.system_name = f['system'].attrs['name'].decode()
            N_atoms = len(atomic_numbers)
            for i, X in enumerate(X_data):
                ind = Individual(cell=cell, numbers=atomic_numbers,
                                 pbc=pbc, positions=X.reshape((N_atoms, 3)))
                ind.my_name = 'Train_individual_' + str(i)
                self.training_population.append(ind)
                # update also new training data, so the model will be
                # initialized
                self.new_training_population.append(ind)
            self.energy_train.extend(list(y_data))
            self.forces_train.append(dy_data)
            self.new_energy_train.extend(list(y_data))
            self.new_forces_train.append(dy_data)

    def update_dataset(self):
        if len(self.new_energy_std_train) > 0:
            energy_std = 3*1000*np.array(self.new_energy_std_train)/self.d
        if self.proc_rank == 0:
            names = [self.population[i].my_name
                     for i in self.new_training_population_indices]
            names = '\t'.join(names)
            c1 = len(self.new_energy_std_train) > 0
            c2 = len(self.new_training_population) > 0
            if c1 and c2 and self.is_fitted:  # fitted with predictions
                logger.info(f'Individuals: {names} will be added to the '
                            'training set. The uncertainties on their '
                            f'energies are: {energy_std} meV/atom')
            elif (not self.is_fitted) and c2:  # not fitted
                logger.info(f'Individuals: {names} will be added to the '
                            'training set. These are the first ones to '
                            'be added')
            else:  # fitted with predictions, but no new training data
                if self.current_generation < self.min_generation:
                    logger.info('Predictions using metamodel will start '
                                f'after generation {self.min_generation}')
                    energies = np.array([x.get_total_energy()
                                         for x in self.population])
                    forces = np.vstack([x.get_forces().ravel()
                                        for x in self.population])                    
                    write_train_hdf5(
                        self.dataset, self.population, energies, forces,
                        self.system_name)
                else:
                    logger.info('All individuals were predicted using the '
                                'metamodel. Nothing to add to the '
                                'training data')

            if c2:
                write_train_hdf5(
                    self.dataset, self.new_training_population,
                    self.new_energy_train, self.new_forces_train,
                    self.system_name)

    def train_model(self):
        # retrain only if new instances are added
        if len(self.new_training_population) > 0:
            if self.proc_rank == 0:
                logger.info('Training the metamodel with '
                            f'{len(self.training_population)} instances')
            self.is_fitted = True
            self.metamodel.fit(self.training_population,
                               np.array(self.energy_train),
                               **self.train_kwargs)
            if self.proc_rank == 0:
                logger.info('Metamodel has been trained.')
        # in any case write the descriptors
        else:
            if self.proc_rank == 0:
                self.metamodel.descriptors_database.write_descriptors(
                    self.population) 

    def _local_predict(self):
        """ Local process has to predict energies for its population and
        eventually decide which individuals should be added to the training
        set
        """
        local_pop = self.population[self.p_start:self.p_end + 1]
        if self.is_fitted:
            logger.info(f'Local predictions for process {self.proc_rank}')
            y_pred, y_stds = self.metamodel.predict(local_pop)
            energies = y_pred
            ene_stds = y_stds
            # sometimes gpytorch predicts nan
            ene_stds[np.isnan(ene_stds)] = np.inf
            local_train_bool = 3*ene_stds/self.d >=\
                               self.metamodel.std_value
            logger.info(f'Done local predictions for process {self.proc_rank}')
        else:
            energies = []
            ene_stds = []
            local_train_bool = np.ones(len(local_pop), dtype=bool)
        # This quantity is used only locally
        self.local_valid_indices = np.where(~local_train_bool)[0]

        self.new_local_training_population_indices = \
            np.where(local_train_bool)[0] + self.p_start
        self.local_predicted_energies = energies
        self.local_predicted_energies_std = ene_stds

    def _get_current_training_indices(self):
        # get the indices of the individuals that will be added to
        # the training set plus other info. For getting the training
        # energies, we need that get_fitness is called first

        # this will be empty if no predictions were made
        stds_pred = self.local_predicted_energies_std
        new_indices = self.new_local_training_population_indices
        indices_start = np.array(new_indices) - self.p_start
        if self.proc_rank == 0:
            self.new_training_population_indices.extend(list(new_indices))
            if len(stds_pred) > 0:
                self.new_energy_std_train.extend(
                    list(stds_pred[indices_start]))
            for j in range(1, self.comm_size):
                no_new_indices = self.comm.recv(source=j, tag=11)
                other_indices = np.emtpy(no_new_indices, dtype=np.uint16)
                self.comm.Recv([other_indices, MPI.LONG_INT],
                               source=j, tag=12)
                other_stds = np.empty(no_new_indices, dtype=DTYPE)
                self.comm.Recv([other_stds, MPI.DOUBLE], source=j, tag=13)
                self.new_training_population_indices.extend(
                    list(other_indices))
                self.new_energy_std_train.extend(list(other_stds))
        else:
            self.comm.send(len(new_indices), dest=0, tag=11)
            self.comm.Send([new_indices, MPI.LONG_INT], dest=0, tag=12)
            self.comm.Send(
                [stds_pred[indices_start], MPI.DOUBLE], dest=0, tag=13)
        self.comm.Barrier()
        if self.proc_rank == 0:
            all_new_training_indices = self.new_training_population_indices
        else:
            all_new_training_indices = None
        all_new_training_indices = self.comm.bcast(all_new_training_indices,
                                                   root=0)
        self.new_training_population_indices = all_new_training_indices

    def get_fitness(self):
        self.p_start, self.p_end = self._get_indices_to_process()
        self.current_generation += 1
        if self.current_generation > self.min_generation:
            self._local_predict()
            self.comm.Barrier()
        # update individuals
        start, end = self.p_start, self.p_end
        if self.proc_rank == 0:
            logger.debug(f'The FitnessCalculator in process {self.proc_rank} '
                         f'of {self.comm_size} will calculate the fitness '
                         f'of individuals from {start} to {end}')
            for j in range(1, self.comm_size):
                r_start = self.comm.recv(source=j, tag=32)
                r_end = self.comm.recv(source=j, tag=33)
                logger.debug(f'The FitnessCalculator in process {j} '
                             f'of {self.comm_size} will calculate the fitness '
                             f'of individuals from {r_start} to {r_end}')
        else:
            self.comm.send(start, dest=0, tag=32)
            self.comm.send(end, dest=0, tag=33)

        local_pop = self.population[start:end + 1]
        my_fitnesses = []
        for i in range(start, end + 1):
            if i in self.local_valid_indices:
                ene = self.local_predicted_energies[i]
                self.population[start + i].calc.calculator_results['energy'] =\
                    ene
                self.population[start + i].skip_calc = True
                my_fitnesses.append(-ene)
                std = self.local_predicted_energies_std[i]
                if self.proc_rank == 0:
                    logger.debug(f'Individual {local_pop[i].my_name} fitness '
                                 f'predicted with the metamodel. '
                                 f'STD: {3*1000*std/self.d} meV/atom. '
                                 f'Predicted energy: {ene}')
                    for j in range(1, self.comm_size):
                        other_name = self.comm.recv(source=j, tag=32)
                        other_std = self.comm.recv(source=j, tag=33)
                        other_ene = self.comm.recv(source=j, tag=34)
                        logger.debug(f'Individual {other_name} fitness '
                                     f'predicted with the metamodel. '
                                     f'STD: {3*1000*other_std/self.d} '
                                     'meV/atom. '
                                     f'Predicted energy: {other_ene}')
                else:
                    self.comm.send(local_pop[i].my_name, dest=0, tag=32)
                    self.comm.send(std, dest=0, tag=33)
                    self.comm.send(ene, dest=0, tag=34)
            else:
                ind = self.population[start + i]
                ene = -ind.fitness
                my_fitnesses.append(-ene)
                # add data for training set
                self.new_local_energy_train.append(ene)
                self.new_local_forces_train.append(ind.get_forces().ravel())

        self.comm.Barrier()
        fitnesses = self._gather_fitnesses(my_fitnesses)
        self.fitnesses = np.array(fitnesses)
        self._gather_training_data()
        self.update_dataset()
        self.train_model()
        self.reset_counts()
        return np.array(fitnesses)

    def _gather_training_data(self):
        # population
        self._get_current_training_indices()
        for index in self.new_training_population_indices:
            self.training_population.append(self.population[index])
            self.new_training_population.append(self.population[index])
        # energies and forces
        if self.proc_rank == 0:
            self.new_energy_train.extend(list(self.new_local_energy_train))
            self.new_forces_train.append(self.new_local_forces_train)
            for j in range(1, self.comm_size):
                no_new_indices = self.comm.recv(source=j, tag=41)
                other_energies = np.emtpy(no_new_indices, dtype=DTYPE)
                self.comm.Recv([other_energies, MPI.DOUBLE],
                               source=j, tag=42)
                other_forces = np.empty((no_new_indices, self.d),
                                        dtype=DTYPE)
                self.comm.Recv([other_forces, MPI.DOUBLE], source=j, tag=43)
                self.new_energy_train.extend(list(other_energies))
                self.new_forces_train.append(other_forces)
        else:
            self.comm.send(len(self.new_local_training_population_indices),
                           dest=0, tag=41)
            energies = np.array(self.new_local_energy_train, dtype=DTYPE)
            self.comm.Send([energies, MPI.DOUBLE], dest=0, tag=42)
            forces = np.array(self.new_local_forces_train, dtype=DTYPE)
            forces = np.vstack(forces)
            self.comm.Send([forces, MPI.DOUBLE], dest=0, tag=43)
        self.comm.Barrier()
        self.new_energy_train = self.comm.bcast(self.new_energy_train, root=0)
        self.new_forces_train = self.comm.bcast(self.new_forces_train, root=0)
        self.comm.Barrier()

        self.energy_train.extend(self.new_energy_train)
        self.forces_train.extend(self.new_forces_train)

        self.new_energy_train = np.array(self.new_energy_train)
        self.new_forces_train = np.vstack(self.new_forces_train)




# -*- coding: utf-8 -*-
""" Copyright 2020 Marco Arrigoni

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import ABC, abstractmethod
import copy
import numpy as np
from scipy.linalg import lstsq
from mpi4py import MPI

import torch
import gpytorch

from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from clinamen.utils import get_structure_id
from clinamen.evpd import ROOT_LOGGER as logger
from .gp_models import ExactGPModel


comm = MPI.COMM_WORLD
proc_no = comm.Get_rank()


class GPMetaModel(ABC):
    """ Class for creating a Gaussian Process metamodel.
    """
    def __init__(self, descriptors_database, 
                 mean_function=None, mean_function_kwargs=None,
                 kernel_function=None, kernel_function_kwargs=None,
                 likelihood_function=None, likelihood_function_kwargs=None,
                 optimizer=None, optimizer_kwargs=None,
                 marginal_likelihood_function=None,
                 marginal_likelihood_function_kwargs=None,
                 preprocessing_pipeline=None, std_value=1e-2):
        """
        Parameters
        ----------
        descriptors_database : object
            Instance of ``DescriptorsDatabase`` class used to produce
            some descriptors.

        mean_function : ``gpytorch.means`` class
            the class that once initialized produces the mean function.
            For example: ``mean_function=gpytorch.means.LinearMean``

        mean_function_kwargs : dict
            the parameters to instantiate the ``mean_function`` class.
            For example: ``mean_function_kwargs={'bias': False}``.
            The actual mean function to be used in the GP model will
            then be the object ``mean_function(**mean_function_kwargs)``

        kernel_function : ``gpytorch.kernels`` class
            the class that once initialized produces the kernel function

        kernel_function_kwargs : dict
            the parameters to instantiate the ``kernel_function`` class

        likelihood_function : ``gpytorch.likelihoods`` class
            the class that once initialized produces the likelihood function

        likelihood_function_kwargs : dict
            the parameters to instantiate the ``likelihood_function`` class

        optimizer : ``torch.optim`` class
            the class that once initialized produces the optimizer for
            maximizing the marginal likelihood

        optimizer_kwargs : dict
           the parameters for initializing the optimizer

        marginal_likelihood_function : ``gpytorch.mlls`` class
            the marginal likelihood class to be maximized

        marginal_likelihood_function_kwargs : dict
            the parameters for initializing the 
            ``marginal_likelihood_function``

        preprocessing_pipeline : ``sklearn.pipeline.Pipeline`` object
            A pipeline to preprocess the inputs data. The metamodel
            is trained using the inputs outputted by this pipeline.
            The trasformation of ``inputs`` must be accomplished by
            calling ``preprocessing_pipeline.fit_transform(``inputs``)``

        std_value : float. Default 1e-2 eV/atom
            when the prediction standard deviation is smaller than
            ``std_value``, the fitness for the individual will be taken
            from the surrogate model. Otherwise, they will be calculated
        """
        self.descriptors_database = descriptors_database
        self._mean_class = mean_function
        self._mean_kwargs = mean_function_kwargs
        self._kernel_class = kernel_function
        self._kernel_kwargs = kernel_function_kwargs
        self._likelihood_class = likelihood_function
        self._likelihood_kwargs = likelihood_function_kwargs
        self._optimizer_class = optimizer
        self._optimizer_kwargs = optimizer_kwargs
        self._mll_class = marginal_likelihood_function
        self._mll_kwargs = marginal_likelihood_function_kwargs
        self.preprocessing_pipeline = preprocessing_pipeline
        self.std_value = std_value

        self.train = True  # If True, we are in training mode
        self._loaded_state = False

    @abstractmethod
    def initialize_model(self, X_train, y_train):
        """ This function initializes the GP model (``gpytorch.models``)
        class, which means it initializes the mean function, kernel function,
        and the likelihood. The function also initializes the optimizer
        and the marginal log likelihood.

        All these initialized objects must be assigned
        to the respectie attributes:

        - ``self._mean_function``
        - ``self._kernel_function``
        - ``self._likelihood``
        - ``self._model``
        - ``self._optimizer``
        - ``self._mll``

        which can then be accessed through the corresponding property
        """

    @property
    def mean_function(self):
        return self._mean_function

    @property
    def kernel_function(self):
        return self._kernel_function

    @property
    def likelihood(self):
        return self._likelihood

    @property
    def model(self):
        return self._model

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def mll(self):
        return self._mll

    @property
    def loaded_state(self):
        """ If True, it means that the state of the model has been
        loaded from an external file
        """
        return self._loaded_state

    def _write_and_retrieve_descriptors(self, structures):
        ### get the descriptors
        ids = [self.descriptors_database.get_structure_id(atom)
               for atom in structures]
        # we write and read from disk every time.
        self.descriptors_database.write_descriptors(structures) 
        descriptors, _ = self.descriptors_database.read_descriptors(ids)
        if proc_no == 0:
            logger.info(
                f'Obtained descriptors of shape {descriptors.shape}')
        ### pass the descriptors through the preprocessing pipeline
        if self.preprocessing_pipeline is None:
            X = descriptors
        else:
            if self.train:
                X = self.preprocessing_pipeline.fit_transform(descriptors)
            else:
                X = self.preprocessing_pipeline.transform(descriptors)
            if proc_no == 0:
                logger.info(
                    f'Transformed descriptors of shape {X.shape}')
        return X

    def fit(self, structures, y, epochs=10000, stopping=1e-3,
            stopping_epochs=10, verbose=False):
        """ Train the meta-model

        Parameters
        ----------
        population : Iterable of structures of length n_samples.

        y : ``np.ndarray`` of shape (n_samples, )
            the total energy of the structures in ``structures``

        epochs : int
            the number of epochs for training the metamodel

        stopping : float
            the loss function minimum change to trigger early stopping

        stopping_epochs : float
            for how many epochs the loss function should change by less than
            ``stopping`` in order to enforce early stopping

        verbose : bool
            If True, prints the loss function every 100 epochs
        """
        if proc_no == 0:
            logger.info(
                f'Training the metamodel with {len(structures)} structures')
        self.train = True
        X_train = self._write_and_retrieve_descriptors(structures)

        # to torch tensors
        X_train = torch.as_tensor(X_train, dtype=torch.float32)
        y = torch.as_tensor(y, dtype=torch.float32)

        ### Initialize all parts of the GP model
        self.initialize_model(X_train, y)

        ### train the metamodel
        self.model.train()
        self.likelihood.train()

        prev_loss_val = np.inf
        iter_stop = 0
        for i in range(epochs):
            self.optimizer.zero_grad()
            output = self.model(X_train)
            loss = -self.mll(output, y)
            loss.backward()
            loss_val = loss.detach().numpy()
            if i == 0:
                if proc_no == 0:
                    logger.info(f'Initial loss: {loss_val:.3f}')
            if verbose and i%100 == 0:
                if proc_no == 0:
                    logger.debug(f'Iter {i+1}/{epochs}.  '
                                 f'Loss = {loss_val:.3f}')
            self.optimizer.step()
            if np.abs(prev_loss_val - loss_val) < stopping:
                iter_stop += 1
                if iter_stop >= stopping_epochs:
                    if proc_no == 0:
                        logger.debug('Early stopping!')
                    break
            else:
                iter_stop = 0
            prev_loss_val = loss_val
        if proc_no == 0:
            logger.info(f'Model trained after {i+1} epochs. '
                        f'Final loss: {loss_val:.3f}')

    def predict(self, structures):
        """ Predict the total energy for each individuals in
        an iterable of structures.

        Parameters
        ----------
        population : Iterable of structures of length n_samples.

        Returns:
        --------
        mean : np.array
            the predicted energies

        std : np.array
            the predicted standard deviations
        """
        if proc_no == 0:
            logger.info(
                f'Making predictions with the metamodel for {len(structures)} '
                 'structures') 
        self.train = False
        X_test = self._write_and_retrieve_descriptors(structures)

        # to torch tensors
        X_test = torch.as_tensor(X_test, dtype=torch.float32)

        ### making predictions
        self.model.eval()
        self.likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            prediction_dist = self.likelihood(self.model(X_test))
            mean = prediction_dist.mean.detach().numpy()
            std = prediction_dist.stddev.detach().numpy()
        return mean, std

    def write_descriptors(self, structures):
        """ Save the descriptors into the database hdf5 file.
        """
        self.descriptors_database.write_descriptors(structures)

    def read_descriptors(self):
        """ Read the descriptors from the hdf5 database file.
        """
        descriptors, _ = self.descriptors_database.read_descriptors()
        return descriptors

    def save_model(self, filename='model_state.pth'):
        """ Save the GP model """
        torch.save(self.model.state_dict(), filename)

    def load_model_state(self, model_pth):
        self.state_dict = torch.load(model_pth)
        self._loaded_state = True


class ExactGPMetaModel(GPMetaModel):
    """ Class for making a meta-model based on a Gaussian Process
    Regressor with exact inference.
    """
    def initialize_model(self, X_train, y_train):
        self._mean_function = self._mean_class(**self._mean_kwargs)
        self._kernel_function = self._kernel_class(
            ard_num_dims=X_train.shape[-1], **self._kernel_kwargs)
        self._likelihood = self._likelihood_class(**self._likelihood_kwargs)
        self._model = ExactGPModel(X_train, y_train, self._mean_function,
                                   self._kernel_function, self._likelihood,
                                   scale_kernel_kwargs=dict())
        self._optimizer = self._optimizer_class(self._model.parameters(),
                                                **self._optimizer_kwargs)
        self._mll = self._mll_class(self._likelihood, self._model,
                                    **self._mll_kwargs)


class BaseExactGPMetaModel(ExactGPMetaModel):
    """ Basic Exact GP regressor with a RBF kernel for minimal initialization
    effort.
    """
    def __init__(self, descriptors_database,
                 preprocessing_pipeline=None, std_value=1e-2):
        """
        Parameters
        ----------
        descriptors_database : object
            Instance of ``DescriptorsDatabase`` class used to produce
            some descriptors.

        preprocessing_pipeline : ``sklearn.pipeline.Pipeline`` object
            A pipeline to preprocess the inputs data. The metamodel
            is trained using the inputs outputted by this pipeline.
            The trasformation of ``inputs`` must be accomplished by
            calling ``preprocessing_pipeline.fit_transform(``inputs``)``

        std_value : float. Default 1e-2 eV/atom
            when the prediction standard deviation is smaller than
            ``std_value``, the fitness for the individual will be taken
            from the surrogate model. Otherwise, they will be calculated
        """
        mean_func = gpytorch.means.ConstantMean
        kernel_func = gpytorch.kernels.RBFKernel
        likelihood = gpytorch.likelihoods.GaussianLikelihood
        noise_constraint = gpytorch.constraints.Interval(1e-5, 1e-1)
        likelihood_kwargs = dict(noise_covar=1e-3, noise_constraint=noise_constraint)
        optimizer = torch.optim.Adam
        optimizer_kwargs = dict(lr=1e-2, amsgrad=True)
        marginal_log_likelihood = gpytorch.mlls.ExactMarginalLogLikelihood

        super().__init__(descriptors_database, mean_func, dict(),
                         kernel_func, dict(), likelihood, likelihood_kwargs,
                         optimizer, optimizer_kwargs,
                         marginal_log_likelihood, dict(),
                         preprocessing_pipeline, std_value)


class BasePCAExactGPMetaModel(BaseExactGPMetaModel):
    """ Basic Exact GP regressor with a RBF kernel for minimal initialization
    effort. The inputs are automatically passed through a pipeline that scales
    them and then performs PCA
    """
    def __init__(self, descriptors_database,
                 scaler_kwargs, pca_kwargs,
                 std_value=1e-2):
        """
        Parameters
        ----------
        descriptors_database : object
            Instance of ``DescriptorsDatabase`` class used to produce
            some descriptors.

        scaler_kwargs : dict
            parameters to initialize a ``sklean`` ``StandardScaler`` object

        pca_kwargs : dict
            parameters to initialize a ``sklearn`` ``PCA`` object

        std_value : float. Default 1e-2 eV/atom
            when the prediction standard deviation is smaller than
            ``std_value``, the fitness for the individual will be taken
            from the surrogate model. Otherwise, they will be calculated
        """
        scaler = StandardScaler(**scaler_kwargs)
        pca = PCA(**pca_kwargs)
        steps = [
            ('scaler', scaler),
            ('pca', pca)
        ]
        pipeline = Pipeline(steps)
        super().__init__(descriptors_database, pipeline, std_value)
